from pathlib import Path

# Not clean, but we import the _PluginManager class instead of using the singleton so we
# have isolation between tests.
from tclint.commands.plugins import _PluginManager as PluginManager

MY_DIR = Path(__file__).parent.resolve()
TEST_DATA_DIR = MY_DIR / "data"


def test_load_openroad():
    plugins = PluginManager()
    loaded = plugins.load_from_spec(TEST_DATA_DIR / "openroad.json")
    assert isinstance(loaded, dict)


def test_load_invalid():
    plugins = PluginManager()
    loaded = plugins.load_from_spec(TEST_DATA_DIR / "invalid.json")
    assert loaded is None



import argparse
import pathlib

import pytest

from tclint.config import (
    get_config,
    RunConfig,
    ConfigError,
    setup_config_cli_args,
    setup_tclfmt_config_cli_args,
)
from tclint.violations import Rule

MY_DIR = pathlib.Path(__file__).parent.resolve()


def test_example_config():
    config_path = MY_DIR / "data" / "tclint.toml"
    config = get_config(config_path, pathlib.Path.cwd())

    global_ = config.get_for_path(pathlib.Path())

    assert global_.exclude == ["ignore_me/", "ignore*.tcl", "/ignore_from_here"]
    assert global_.ignore == [Rule("unbraced-expr")]
    assert global_.extensions == ["tcl"]
    assert global_.commands == pathlib.Path("~/.tclint/openroad.json")

    assert global_.style_indent == 2
    assert global_.style_line_length == 80
    assert global_.style_indent_namespace_eval is False
    assert global_.style_spaces_in_braces is True

    group1 = config.get_for_path(pathlib.Path("other_file_group1/file.tcl"))
    assert group1.style_indent == 3
    assert group1.ignore == [Rule("command-args")]
    assert group1.style_line_length == 80

    group2 = config.get_for_path(pathlib.Path("other_file_group2/foo/file.tcl"))
    assert group2.style_indent == 2
    assert group2.style_spaces_in_braces is False


def test_invalid_rule():
    with pytest.raises(ConfigError) as excinfo:
        RunConfig.from_dict({"ignore": ["asdf"]}, pathlib.Path.cwd())

    if excinfo is not None:
        print(str(excinfo.value))


def test_pyproject():
    config = RunConfig.from_pyproject(MY_DIR / "data")
    global_ = config.get_for_path(pathlib.Path())
    assert global_.style_indent == 2
    assert global_.ignore == [Rule("unbraced-expr")]


def test_tclint_config_args():
    parser = argparse.ArgumentParser("tclint")
    setup_config_cli_args(parser)

    args = [
        "--ignore",
        "unbraced-expr, line-length",
        "--extend-ignore",
        "trailing-whitespace",
        "--exclude",
        "my_dir,my_file",
        "--extend-exclude",
        "extend_to_file",
        "--extensions",
        "sdc, exp",
        "--commands",
        "commands.json",
        "--style-line-length",
        "79",
    ]

    args = parser.parse_args(args)

    assert args.ignore == [Rule("unbraced-expr"), Rule("line-length")]
    assert args.extend_ignore == [Rule("trailing-whitespace")]
    assert args.exclude == ["my_dir", "my_file"]
    assert args.extend_exclude == ["extend_to_file"]
    assert args.extensions == ["sdc", "exp"]
    assert args.commands == pathlib.Path("commands.json")
    assert args.style_line_length == 79


def test_tclfmt_config_args():
    parser = argparse.ArgumentParser("tclfmt")
    setup_tclfmt_config_cli_args(parser)

    args = [
        "--exclude",
        "my_dir,my_file",
        "--extend-exclude",
        "extend_to_file",
        "--extensions",
        "sdc, exp",
        "--commands",
        "commands.json",
        "--indent",
        "5",
        "--max-blank-lines",
        "4",
        "--indent-namespace-eval",
        "--no-spaces-in-braces",
    ]

    args = parser.parse_args(args)

    assert args.exclude == ["my_dir", "my_file"]
    assert args.extend_exclude == ["extend_to_file"]
    assert args.extensions == ["sdc", "exp"]
    assert args.commands == pathlib.Path("commands.json")
    assert args.style_indent == 5
    assert args.style_max_blank_lines == 4
    assert args.style_indent_namespace_eval is True
    assert args.style_spaces_in_braces is False


def test_invalid_tclint_args():
    parser = argparse.ArgumentParser("tclint", exit_on_error=False)
    setup_config_cli_args(parser)

    for arg, value in [
        ("--ignore", "unbraced-expr, line-length, invalid-rule"),
        ("--extend-ignore", "trailing-whitespace, invalid-rule"),
        ("--style-line-length", "asdf"),
    ]:
        with pytest.raises(argparse.ArgumentError) as excinfo:
            parser.parse_args([arg, value])

        # For manually auditing error messages
        print(excinfo.value)


def test_invalid_tclfmt_args():
    parser = argparse.ArgumentParser("tclfmt", exit_on_error=False)
    setup_tclfmt_config_cli_args(parser)

    for arg, value in [
        ("--indent", "qwerty"),
        ("--max-blank-lines", "uiop"),
    ]:
        with pytest.raises(argparse.ArgumentError) as excinfo:
            parser.parse_args([arg, value])

        # For manually auditing error messages
        print(excinfo.value)


def test_invalid_config():
    for config in [
        {"ignore": ["unbraced-expr", "invalid-rule", "line-length"]},
        {"commands": 1},
        {"style": {"indent": "abc"}},
        {"style": {"line-length": "def"}},
        {"style": {"max-blank-lines": "ghi"}},
        {"style": {"indent-namespace-eval": "not-a-bool"}},
        {"style": {"spaces-in-braces": -1}},
        {"fileset": [{"style": {"indent": 4}}]},  # missing key
        {"fileset": [{"paths": ["dir"], "style": {"indent": "not-a-number"}}]},
        {"extra-key": "value"},
    ]:
        with pytest.raises(ConfigError) as excinfo:
            RunConfig.from_dict(config, pathlib.Path.cwd())

        # For manually auditing error messages
        print(excinfo.value)



import json


def main():
    new = {}
    with open("tests/commands/data/openroad.json", "r") as f:
        data = json.load(f)

    new["name"] = data["name"]
    new["commands"] = {}
    for command, spec in data["commands"].items():
        new["commands"][command] = {}

        min_ = spec["positionals"]["min"]
        max_ = spec["positionals"]["max"]
        positionals = []
        for i in range(min_):
            positionals.append({
                "name": f"arg{i + 1}",
                "required": True,
                "value": {"type": "any"},
            })

        if max_ is None:
            positionals.append({
                "name": "remaining",
                "required": False,
                "value": {"type": "variadic"},
            })
        else:
            for i in range(min_, max_):
                positionals.append({
                    "name": f"arg{i + 1}",
                    "required": False,
                    "value": {"type": "any"},
                })

        if positionals:
            new["commands"][command]["positionals"] = positionals

        switches = {}
        for switch, switch_spec in spec["switches"].items():
            switches[switch] = {
                "required": switch_spec["required"],
                "value": {"type": "any"} if switch_spec["value"] else None,
                "repeated": switch_spec["repeated"],
            }

        if switches:
            new["commands"][command]["switches"] = switches

    with open("tests/commands/data/openroad.json", "w") as f:
        json.dump(new, f)


if __name__ == "__main__":
    main()



from pathlib import Path
import sys

import pytest

from tclint.cli.utils import make_exclude_filter


# These are not regularly run in CI, but were run at time of check-in. Keeping them
# around for reference.
@pytest.mark.skipif(
    sys.platform != "win32", reason="testing edge cases with Windows path handling"
)
@pytest.mark.parametrize(
    "pattern,path,root,excluded",
    [
        ("foo.tcl", "C:\\foo.tcl", "D:\\", True),
        ("foo.tcl", "C:\\blah.tcl", "D:\\", False),
        ("../foo.tcl", "C:\\foo.tcl", "D:\\", False),
        ("/foo.tcl", "C:\\foo.tcl", "D:\\", False),
    ],
)
def test_exclude_filter_windows(pattern, path, root, excluded):
    is_excluded = make_exclude_filter([pattern], Path(root))
    assert is_excluded(Path(path)) == excluded



"""Classes for representing and interacting with Tcl syntax trees. """


class Visitor:
    """Abstract base class for Visitors that operate on syntax tree."""

    def visit_script(self, script):
        pass

    def visit_comment(self, comment):
        pass

    def visit_command(self, command):
        pass

    def visit_command_sub(self, command_sub):
        pass

    def visit_bare_word(self, word):
        pass

    def visit_braced_word(self, word):
        pass

    def visit_quoted_word(self, word):
        pass

    def visit_compound_bare_word(self, word):
        pass

    def visit_var_sub(self, var_sub):
        pass

    def visit_arg_expansion(self, arg_expansion):
        pass

    def visit_list(self, list):
        pass

    def visit_expression(self, expression):
        pass

    def visit_braced_expression(self, expression):
        pass

    def visit_paren_expression(self, expression):
        pass

    def visit_unary_op(self, unary_op):
        pass

    def visit_binary_op(self, binary_op):
        pass

    def visit_ternary_op(self, ternary_op):
        pass

    def visit_function(self, function):
        pass


class Node:
    """
    Invariants:
    - self.value is some sort of base Python type
    - self.children is a list of Node types
    """

    def __init__(self, *init, pos=None, end_pos=None):
        """pos: line, column of first character of parsed region (1-indexed)
        end_pos: line, column of first character after parsed region (1-indexed)
        """
        self.line = None
        self.col = None
        if pos is not None:
            self.line, self.col = pos
        self.end_pos = end_pos

        self.value = None
        if len(init) > 0 and not isinstance(init[0], Node):
            self.value = init[0]
            init = init[1:]

        if not all(isinstance(v, Node) for v in init):
            raise TypeError("Children must be Node instances")

        self.children = list(init)

    def add(self, node):
        self.children.append(node)

    @property
    def contents(self):
        """This is overloaded by word Nodes that may have concrete contents.

        TODO: I prefer the name value, but that's currently taken...
        """
        return None

    @property
    def contents_pos(self):
        """This is overloaded by word Nodes that may have concrete contents.

        Returns the position at which the contents start.

        TODO: consider combining with with `contents`?
        """
        return None

    def _pos_str(self):
        start_pos_str = "?"
        if self.pos is not None:
            start_pos_str = f"{self.pos[0]}:{self.pos[1]}"
        end_pos_str = "?"
        if self.end_pos is not None:
            end_pos_str = f"{self.end_pos[0]}:{self.end_pos[1]}"

        return f"  # {start_pos_str}-{end_pos_str}"

    def _make_str(self, indent=None, positions=False):
        if indent is not None:
            s = "  " * indent
        else:
            s = ""

        s += self.__class__.__name__
        s += "("

        if self.value:
            s += repr(self.value)
            if self.children:
                s += ", "

        if positions and self.children:
            s += self._pos_str()

        for i, child in enumerate(self.children):
            if indent is not None:
                s += "\n"
            s += child._make_str(
                indent=None if indent is None else indent + 1, positions=positions
            )
            if i < len(self.children) - 1:
                s += ", "
        s += ")"

        if positions and not self.children:
            s += self._pos_str()

        return s

    def pretty(self, positions=False):
        return self._make_str(indent=0, positions=positions)

    def __str__(self):
        return self._make_str()

    def __eq__(self, other):
        if type(self) is not type(other):
            return False

        if self.value != other.value:
            return False

        if len(self.children) != len(other.children):
            return False
        for my_child, other_child in zip(self.children, other.children):
            if my_child != other_child:
                return False

        return True

    def diff(self, other, indent_depth=0):
        lines = []
        indent = "  " * indent_depth

        my_cls = self.__class__.__name__
        other_cls = other.__class__.__name__

        if my_cls != other_cls:
            lines += [f"{indent}-{my_cls}("]
            lines += [f"{indent}+{other_cls}("]
            return lines

        if self.value != other.value:
            lines += [f'{indent}-{my_cls}("{self.value}"']
            lines += [f'{indent}+{other_cls}("{other.value}"']
            return lines

        if len(self.children) != len(other.children):
            my_children = ",".join([
                child.__class__.__name__ for child in self.children
            ])
            other_children = ",".join([
                child.__class__.__name__ for child in other.children
            ])

            lines += [f"{indent}-{my_cls}({my_children})"]
            lines += [f"{indent}+{other_cls}({other_children})"]
            return lines

        if self.value is not None:
            lines += [f"{indent}{my_cls}({self.value}"]
        else:
            lines += [f"{indent}{my_cls}("]

        for my_child, other_child in zip(self.children, other.children):
            lines += my_child.diff(other_child, indent_depth=indent_depth + 1)

        lines += [f"{indent})"]

        return lines

    @property
    def pos(self):
        if self.line is None or self.col is None:
            return None

        return (self.line, self.col)

    def _recurse(self, visitor):
        for child in self.children:
            child.accept(visitor, recurse=True)


class Script(Node):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # hack for spaces-in-braces check
        self.braced = False

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_script(self)


class Comment(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_comment(self)


class Command(Node):
    def __init__(self, routine, *args, pos=None, end_pos=None):
        self.routine = routine
        self.args = args
        super().__init__(routine, *args, pos=pos, end_pos=end_pos)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_command(self)


class CommandSub(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_command_sub(self)


class BareWord(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_bare_word(self)

    @property
    def contents(self):
        return self.value

    @property
    def contents_pos(self):
        return self.pos


class BracedWord(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_braced_word(self)

    @property
    def contents(self):
        return self.value

    @property
    def contents_pos(self):
        return (self.line, self.col + 1)


class QuotedWord(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_quoted_word(self)

    @property
    def contents(self):
        """A QuotedWord with concrete contents is put in the tree as
        QuotedWord(BareWord()), so return contents of its child."""
        if len(self.children) > 1:
            return None
        if len(self.children) == 0:
            # weird special case to handle blank quoted string ("") - it doesn't
            # make sense to put in a child, but setting/returning the value is
            # also inconsistent
            return ""

        return self.children[0].contents

    @property
    def contents_pos(self):
        if self.contents is None:
            return None
        return (self.line, self.col + 1)


class CompoundBareWord(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_compound_bare_word(self)


class VarSub(Node):
    def __init__(self, *args, braced=False, **kwargs):
        self.braced = braced
        return super().__init__(*args, **kwargs)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_var_sub(self)


class ArgExpansion(Node):
    def __init__(self, list, pos=None, end_pos=None):
        self.list = list
        super().__init__(list, pos=pos, end_pos=end_pos)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_arg_expansion(self)


class List(Node):
    """This Node currently exists exclusively for implementing the switch
    command in a way that facilitates style checks. Might be nice to find
    another way to handle this that doesn't require a special Node."""

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_list(self)


class Expression(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_expression(self)


class BracedExpression(Node):
    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_braced_expression(self)


class ParenExpression(Node):
    def __init__(self, body: Expression, pos=None, end_pos=None):
        self.body = body
        super().__init__(body, pos=pos, end_pos=end_pos)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_paren_expression(self)


class UnaryOp(Node):
    def __init__(self, operator, operand, pos=None, end_pos=None):
        self.operator = operator
        self.operand = operand
        super().__init__(operator, operand, pos=pos, end_pos=end_pos)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_unary_op(self)


class BinaryOp(Node):
    def __init__(self, operator1, operand, operator2, pos=None, end_pos=None):
        self.operator1 = operator1
        self.operand = operand
        self.operator2 = operator2
        super().__init__(operator1, operand, operator2, pos=pos, end_pos=end_pos)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_binary_op(self)


class TernaryOp(Node):
    def __init__(self, condition, question, true, colon, false, pos=None, end_pos=None):
        self.condition = condition
        self.question = question
        self.true = true
        self.colon = colon
        self.false = false

        super().__init__(
            condition, question, true, colon, false, pos=pos, end_pos=end_pos
        )

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_ternary_op(self)


class Function(Node):
    def __init__(self, name, *args, pos=None, end_pos=None):
        self.name = name
        self.args = args
        super().__init__(name, *args, pos=pos, end_pos=end_pos)

    def accept(self, visitor, recurse=False):
        if recurse:
            self._recurse(visitor)
        visitor.visit_function(self)



"""Helpers for checking command arguments."""

from collections.abc import Callable
from typing import List, Optional, Union

from tclint.syntax_tree import ArgExpansion, QuotedWord, BracedWord, BareWord, Node


class CommandArgError(Exception):
    pass


def arg_count(args, parser):
    # TODO: graceful handling of argsub going into things with recursive parsing.
    # if the argsub happens to be "concrete", we can technically do the right
    # thing (although this should probably be flagged as a readability issue...)
    # otherwise, we should flag that the non-concrete argsub is not okay for
    # these cases. however, I think its not okay-ness doesn't need to be absolute, e.g.
    # I think we could allow:
    #
    #  catch {puts "my script"} {*}$catchopts
    #

    arg_count = 0
    has_arg_expansion = False
    for arg in args:
        if isinstance(arg, ArgExpansion):
            if arg.contents is None:
                has_arg_expansion = True
                continue
            arg_count += len(parser.parse_list(arg.contents))
        else:
            arg_count += 1

    return arg_count, has_arg_expansion


def check_count(command, min=None, max=None, args_name="args"):
    def check(args, parser):
        if min is None and max is None:
            return None

        count, has_arg_expansion = arg_count(args, parser)

        if not has_arg_expansion and min == max and count != min:
            raise CommandArgError(
                f"wrong # of {args_name} for {command}: got {count}, expected {min}"
            )

        if not has_arg_expansion and min is not None and count < min:
            raise CommandArgError(
                f"not enough {args_name} for {command}: got {count}, expected at least"
                f" {min}"
            )

        if max is not None and count > max:
            raise CommandArgError(
                f"too many {args_name} for {command}: got {count}, expected no more"
                f" than {max}"
            )

        return None

    return check


def eval(args, parser, command):
    if len(args) > 1 and any(isinstance(arg, (QuotedWord, BracedWord)) for arg in args):
        # Slightly odd restriction, but our syntax tree doesn't have a great way
        # to handle this case. We require each command argument to correspond to
        # one child node, but multiple quoted or braced word arguments can be
        # combined into a single subcommand when interpreted eval-style. This
        # requirement exists to facilitate style checking, if we had a separate
        # CST for style checks and AST for logical checks we may be able to
        # handle it.

        raise CommandArgError(
            f"unable to parse multiple {command} arguments when one includes a braced"
            " or quoted word"
        )

    # Construct the body of the eval taking whitespace into account to ensure we get
    # style checking.

    eval_script = ""
    prev_arg_end_pos = None
    for arg in args:
        contents = arg.contents
        if contents is None:
            # TODO: flag sort of eval-specific violation? Common patterns will
            # often trigger this, and it seems useful to be able to turn it off
            raise CommandArgError(
                f"{command} received an argument with a substitution, unable to parse"
                " its arguments"
            )

        if prev_arg_end_pos is not None:
            if prev_arg_end_pos[0] != arg.line:
                # If we have multiple args on the same line, we know there must be a
                # backslash newline. Add it so the parsing works.
                eval_script += "\\\n" * (arg.line - prev_arg_end_pos[0])
                eval_script += " " * (arg.col - 1)
            else:
                eval_script += " " * (arg.col - prev_arg_end_pos[1])
        eval_script += contents

        prev_arg_end_pos = arg.end_pos

    script = parser.parse(eval_script, pos=(args[0].pos))
    script.end_pos = args[-1].end_pos

    return [script]


def check_command(
    command: str, args: List[Node], parser, command_spec: Union[Callable, dict, None]
) -> Optional[List[Node]]:
    if command_spec is None:
        return None

    if isinstance(command_spec, dict):
        return check_arg_spec(command, args, parser, command_spec)

    return command_spec(args, parser)


def check_arg_spec(
    command: str, args: List[Node], parser, arg_spec: dict
) -> Optional[List[Node]]:
    if "subcommands" in arg_spec:
        subcommands = arg_spec["subcommands"]
        try:
            subcommand = args[0].contents
        except IndexError:
            subcommand = None

        if subcommand in subcommands:
            new_args = check_command(
                f"{command} {subcommand}", args[1:], parser, subcommands[subcommand]
            )
            if new_args is None:
                return new_args
            return args[0:1] + new_args

        if "" in subcommands:
            return check_command(command, args, parser, subcommands[""])

        if subcommand is not None:
            msg = f"invalid subcommand for {command}: got {subcommand}"
        else:
            msg = f"no subcommand provided for {command}"

        raise CommandArgError(f"{msg}, expected one of {', '.join(subcommands.keys())}")

    switches = arg_spec["switches"]
    args_allowed = set(switches)
    args_required = {switch for switch in switches if switches[switch]["required"]}
    positional_args = []

    args = list(args)
    while len(args) > 0:
        arg = args.pop(0)

        # To facilitate better error messages, we expect that switches are always
        # specified as BareWords that start with "-" or ">". This lets us throw an
        # error when a switch-like thing doesn't match any supported arguments,
        # rather than counting it towards the positional arguments (which usually
        # ends up in a vague "too many arguments" error). To make tclint interpret a
        # switch-like word as a positional argument, users should wrap it in "", and
        # any switches should be BareWords.
        contents = arg.contents
        if not (isinstance(arg, BareWord) and contents and contents[0] in {"-", ">"}):
            positional_args.append(arg)
            continue

        # TODO check required arguments
        if contents in args_allowed:
            if switches[contents]["value"]:
                try:
                    args.pop(0)
                except IndexError:
                    raise CommandArgError(
                        f"invalid arguments for {command}: expected value after"
                        f" {contents}"
                    )
            if not switches[contents]["repeated"]:
                args_allowed.remove(contents)
            if contents in args_required:
                args_required.remove(contents)
        elif contents in arg_spec:
            raise CommandArgError(f"duplicate argument for {command}: {contents}")
        else:
            prefix_matches = []
            for switch in switches:
                if switch.startswith(contents):
                    prefix_matches.append(switch)

            if len(prefix_matches) == 1:
                raise CommandArgError(
                    f"shortened argument for {command}: expand {contents} to"
                    f" {prefix_matches[0]}"
                )

            if len(prefix_matches) > 1:
                raise CommandArgError(
                    f"ambiguous argument for {command}: {contents} could be any of"
                    f" {', '.join(prefix_matches)}"
                )

            raise CommandArgError(f"unrecognized argument for {command}: {contents}")

    if len(args_required) > 1:
        raise CommandArgError(
            f"missing required arguments for {command}: {', '.join(args_required)}"
        )
    elif len(args_required) == 1:
        raise CommandArgError(
            f"missing required argument for {command}: {args_required.pop()}"
        )

    min_positionals = 0
    max_positionals: Optional[int] = 0
    for positional in arg_spec["positionals"]:
        if positional["value"]["type"] == "variadic":
            max_positionals = None

        if positional["required"]:
            min_positionals += 1
        if max_positionals is not None:
            max_positionals += 1

    check = check_count(
        command,
        min=min_positionals,
        max=max_positionals,
        args_name="positional args",
    )
    check(positional_args, None)

    return None



import contextlib
import pytest

from tclint.commands import CommandArgError
from tclint.commands.checks import check_arg_spec
from tclint.syntax_tree import BareWord, ArgExpansion, VarSub, BracedWord


def test_repeated_switch_allowed():
    args = [BareWord("-abc"), BareWord("-abc")]
    spec = {
        "positionals": [],
        "switches": {
            "-abc": {"required": False, "value": None, "repeated": True},
        },
    }
    check_arg_spec("command", args, None, spec)


def test_repeated_switch_not_allowed():
    args = [BareWord("-abc"), BareWord("-abc")]
    spec = {
        "positionals": [],
        "switches": {
            "-abc": {"required": False, "value": None, "repeated": False},
        },
    }
    with pytest.raises(CommandArgError):
        check_arg_spec("command", args, None, spec)


def test_positional_count_too_many():
    args = [BareWord("foo")]
    spec = {"positionals": [], "switches": {}}
    with pytest.raises(CommandArgError):
        check_arg_spec("command", args, None, spec)


def test_positional_count_too_few():
    args = [BareWord("foo")]
    spec = {
        "positionals": [
            {"name": "arg1", "value": {"type": "any"}, "required": True},
            {"name": "arg2", "value": {"type": "any"}, "required": True},
        ],
        "switches": {},
    }
    with pytest.raises(CommandArgError):
        check_arg_spec("command", args, None, spec)


def test_positional_count_unlimited():
    args = [BareWord("foo")] * 10
    spec = {
        "positionals": [
            {"name": "args", "value": {"type": "variadic"}, "required": True}
        ],
        "switches": {},
    }
    check_arg_spec("command", args, None, spec)


def test_positional_count_arg_expansion():
    args = [BareWord("foo"), ArgExpansion(VarSub("foo"))]
    spec = {
        "positionals": [
            {"name": "arg1", "value": {"type": "any"}, "required": True},
            {"name": "arg2", "value": {"type": "any"}, "required": True},
            {"name": "remaining", "value": {"type": "variadic"}, "required": True},
        ],
        "switches": {},
    }
    check_arg_spec("command", args, None, spec)


def test_required_switch():
    spec = {
        "positionals": [],
        "switches": {
            "-foo": {
                "required": True,
                "value": None,
                "repeated": False,
            },
            "-bar": {
                "required": False,
                "value": None,
                "repeated": False,
            },
        },
    }

    check_arg_spec("command", [BareWord("-foo"), BareWord("-bar")], None, spec)
    check_arg_spec("command", [BareWord("-foo")], None, spec)

    with pytest.raises(CommandArgError) as excinfo:
        check_arg_spec("command", [BareWord("-bar")], None, spec)
    if excinfo is not None:
        print(excinfo.value)

    with pytest.raises(CommandArgError) as excinfo:
        check_arg_spec("command", [], None, spec)
    if excinfo is not None:
        print(excinfo.value)


@pytest.mark.parametrize(
    "args,valid",
    [
        ([BareWord("foo"), BareWord("arg1")], True),
        ([BareWord("foo")], False),  # missing required argument
        ([BareWord("bar"), BareWord("-asdf")], True),
        ([BareWord("baz"), BareWord("quz")], True),
        ([BareWord("baz")], False),  # missing subcommand
        ([BareWord("-default"), BareWord("val")], True),
        ([BareWord("arg1")], False),
    ],
)
def test_subcommands(args, valid):
    spec = {
        "subcommands": {
            "foo": {
                "positionals": [
                    {"name": "arg1", "required": True, "value": {"type": "any"}}
                ],
                "switches": {},
            },
            "bar": None,
            "baz": {
                "subcommands": {
                    "quz": None,
                }
            },
            "": {
                "positionals": [],
                "switches": {
                    "-default": {
                        "required": True,
                        "value": {"type": "any"},
                        "repeated": False,
                    }
                },
            },
        }
    }

    if valid:
        cm = contextlib.nullcontext(None)
    else:
        cm = pytest.raises(CommandArgError)
    with cm as excinfo:
        check_arg_spec("command", args, None, spec)

    if excinfo is not None:
        print(excinfo.value)


def test_arg_replacement_subcommands():
    args = [BareWord("subcommand"), BracedWord("arg")]

    def _parse(args, _):
        return [BareWord(args[0].contents)]

    spec = {
        "subcommands": {
            "subcommand": _parse,
        }
    }
    new_args = check_arg_spec("command", args, None, spec)
    assert new_args == [BareWord("subcommand"), BareWord("arg")]



"""Main CLI entry point."""

import argparse
import pathlib
import sys
from typing import Dict, List, Optional


from tclint.config import (
    get_config,
    setup_config_cli_args,
    Config,
    ConfigError,
    RunConfig,
)
from tclint.parser import Parser, TclSyntaxError
from tclint.checks import get_checkers
from tclint.violations import Violation, Rule
from tclint.comments import CommentVisitor
from tclint.cli.utils import resolve_sources, register_codec_warning

try:
    from tclint._version import __version__  # type: ignore
except ModuleNotFoundError:
    __version__ = "(unknown version)"

# exit code flags
EXIT_OK = 0
EXIT_LINT_VIOLATIONS = 1
EXIT_SYNTAX_ERROR = 2
EXIT_INPUT_ERROR = 4


def filter_violations(
    violations: List[Violation],
    config_ignore: List[Rule],
    inline_ignore: Dict[int, List[Rule]],
) -> List[Violation]:
    filtered_violations = []

    for violation in violations:
        if violation.id in config_ignore:
            continue
        line = violation.start[0]
        if line in inline_ignore and violation.id in inline_ignore[line]:
            continue

        filtered_violations.append(violation)

    return filtered_violations


def lint(
    script: str,
    config: Config,
    path: Optional[pathlib.Path],
    debug=0,
) -> List[Violation]:
    plugins = [config.commands] if config.commands is not None else []
    parser = Parser(debug=(debug > 0), command_plugins=plugins)

    violations = []
    tree = parser.parse(script)
    violations += parser.violations

    if debug > 0:
        print(tree.pretty(positions=(debug > 1)))

    for checker in get_checkers():
        violations += checker.check(script, tree, config)

    v = CommentVisitor()
    ignore_lines = v.run(tree, path)
    violations = filter_violations(violations, config.ignore, ignore_lines)

    return violations


def main():
    parser = argparse.ArgumentParser("tclint")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "source",
        nargs="+",
        help="files to lint. Provide '-' to read from stdin",
        type=pathlib.Path,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="count",
        default=0,
        help=(
            "display debug output. Provide additional times to increase the verbosity"
            " of output (e.g. -dd)"
        ),
    )
    parser.add_argument(
        "-c",
        "--config",
        help="path to config file",
        type=pathlib.Path,
        default=None,
        metavar="<path>",
    )
    setup_config_cli_args(parser)
    args = parser.parse_args()

    try:
        config = get_config(args.config, pathlib.Path())
    except ConfigError as e:
        print(f"Invalid config file: {e}")
        return EXIT_INPUT_ERROR

    if config is None:
        config = RunConfig()

    config.apply_cli_args(args)

    try:
        # TODO: we should eventually allow tclint to find a config by walking up
        # directories, at which point exclude_root should be the parent dir of
        # the config file, unless -c is used (eslint rules)
        exclude_root = pathlib.Path.cwd()
        sources = resolve_sources(
            args.source,
            exclude_patterns=config.exclude,
            exclude_root=exclude_root,
            extensions=config.extensions,
        )
    except FileNotFoundError as e:
        print(f"Invalid path provided: {e}")
        return EXIT_INPUT_ERROR

    retcode = EXIT_OK

    register_codec_warning("replace_with_warning")

    for path in sources:
        if path is None:
            script = sys.stdin.read()
            out_prefix = "(stdin)"
        else:
            with open(path, "r", errors="replace_with_warning") as f:
                script = f.read()
            out_prefix = str(path)

        try:
            violations = lint(
                script,
                config.get_for_path(path),
                path,
                debug=args.debug,
            )
        except TclSyntaxError as e:
            line, col = e.start
            print(f"{out_prefix}:{line}:{col}: syntax error: {e}")
            retcode |= EXIT_SYNTAX_ERROR
            continue

        for violation in sorted(violations):
            print(f"{out_prefix}:{violation}")

        if len(violations) > 0:
            retcode |= EXIT_LINT_VIOLATIONS

    return retcode


if __name__ == "__main__":
    sys.exit(main())



import argparse
import dataclasses
import logging
from pathlib import Path
from typing import Dict, List, Optional, Tuple
import uuid

from lsprotocol import types as lsp

from pygls.server import LanguageServer
from pygls.workspace import TextDocument
from pygls.uris import to_fs_path

from tclint.cli import tclint
from tclint.config import get_config, DEFAULT_CONFIGS, RunConfig, Config, ConfigError
from tclint.format import Formatter, FormatterOpts
from tclint.lexer import TclSyntaxError
from tclint.parser import Parser
from tclint.cli import utils

try:
    from tclint._version import __version__  # type: ignore
except ModuleNotFoundError:
    __version__ = "(unknown version)"


DIAGNOSTIC_SOURCE = "tclint"


def lint(source, config, path):
    diagnostics = []

    try:
        violations = tclint.lint(source, config, path)
    except TclSyntaxError as e:
        return [
            lsp.Diagnostic(
                message=str(e),
                severity=lsp.DiagnosticSeverity.Error,
                range=lsp.Range(
                    start=lsp.Position(e.start[0] - 1, e.start[1] - 1),
                    end=lsp.Position(e.end[0] - 1, e.end[1] - 1),
                ),
                code="syntax error",
                source=DIAGNOSTIC_SOURCE,
            )
        ]

    for violation in violations:
        message = violation.message
        severity = lsp.DiagnosticSeverity.Warning
        start = lsp.Position(
            line=violation.start[0] - 1, character=violation.start[1] - 1
        )
        end = lsp.Position(line=violation.end[0] - 1, character=violation.end[1] - 1)

        diagnostics.append(
            lsp.Diagnostic(
                message=message,
                severity=severity,
                range=lsp.Range(
                    start=start,
                    end=end,
                ),
                code=violation.id,
                source=DIAGNOSTIC_SOURCE,
            )
        )

    return diagnostics


@dataclasses.dataclass
class ExtensionSettings:
    # This path is expected to be absolute.
    config_file: Optional[Path] = dataclasses.field(default=None)


class TclspServer(LanguageServer):
    """Main server class. Implements pull diagnostics using a method adapted from
    https://pygls.readthedocs.io/en/latest/examples/pull-diagnostics.html."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.diagnostics = {}
        self.global_config: RunConfig = None
        # Maps workspace roots to configs.
        self.configs: Dict[Path, RunConfig] = {}
        self.client_supports_refresh = False

        self.global_settings = ExtensionSettings()
        self.workspace_settings: Dict[Path, ExtensionSettings] = {}

    def get_roots(self) -> List[Path]:
        """Returns root folders currently open in the workspace."""
        roots = []
        for uri in self.workspace.folders.keys():
            path = to_fs_path(uri)
            if path is not None:
                roots.append(Path(path))

        if len(roots) > 0:
            return roots

        if self.workspace.root_path is not None:
            roots.append(Path(self.workspace.root_path))

        return roots

    def get_root(self, path: Path) -> Optional[Path]:
        """Returns workspace root folder that's closest to path.

        Returns None if path is not in a workspace folder or if there are no workspace
        folders.
        """
        roots = self.get_roots()
        closest_root = None
        distance = float("inf")
        for root in roots:
            try:
                relpath = path.relative_to(root)
            except ValueError:
                continue
            if len(relpath.parts) < distance:
                distance = len(relpath.parts)
                closest_root = root
        return closest_root

    def get_config_file(self, workspace_root: Path) -> Optional[Path]:
        if workspace_root in self.workspace_settings:
            settings = self.workspace_settings[workspace_root]
            return settings.config_file
        return self.global_settings.config_file

    def load_configs(self):
        self.configs = {}
        for root in self.get_roots():
            try:
                path = self.get_config_file(root)
                config = get_config(path, root)
                if config is not None:
                    self.configs[root] = config
            except ConfigError as e:
                self.show_message(f"Error loading config file: {e}")

        # If a global config file exists, we apply it to any file not under a workspace
        # folder.
        global_path = self.global_settings.config_file
        if global_path is not None:
            try:
                config = get_config(global_path, global_path.parent)
                self.global_config = config
            except ConfigError as e:
                self.show_message(f"Error loading config file: {e}")

    def get_config(self, path: Path, root: Optional[Path]) -> Config:
        if root in self.configs:
            return self.configs[root].get_for_path(path)
        if self.global_config is not None:
            return self.global_config.get_for_path(path)
        return Config()

    def _compute_diagnostics(self, document: TextDocument) -> List[lsp.Diagnostic]:
        path = Path(document.path)
        root = self.get_root(path)
        config = self.get_config(path, root)

        if root is None:
            root = path.parent

        is_excluded = utils.make_exclude_filter(config.exclude)
        if is_excluded(path, root):
            return []

        return lint(document.source, config, path)

    def compute_diagnostics(self, document: TextDocument):
        # `None` sentinel ensures that `diagnostics` gets updated if the URI is not
        # present.
        _, previous = self.diagnostics.get(document.uri, (0, None))

        diagnostics = self._compute_diagnostics(document)

        # Only update if the list has changed
        if previous != diagnostics:
            self.diagnostics[document.uri] = (document.version, diagnostics)

    def format(
        self,
        document: TextDocument,
        options: lsp.FormattingOptions,
        range: Optional[Tuple[int, int]] = None,
    ):
        path = Path(document.path)
        root = self.get_root(path)
        config = self.get_config(path, root)

        parser = Parser()

        if config.style_indent is None:
            indent = "\t" if not options.insert_spaces else " " * options.tab_size
        else:
            indent = config.get_indent()

        formatter = Formatter(
            FormatterOpts(
                indent=indent,
                spaces_in_braces=config.style_spaces_in_braces,
                max_blank_lines=config.style_max_blank_lines,
                indent_namespace_eval=config.style_indent_namespace_eval,
            )
        )

        if range is not None:
            start, end = range
            return formatter.format_partial(document.source[start:end], parser)

        return formatter.format_top(document.source, parser)


server = TclspServer("tclsp", __version__)


@server.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: TclspServer, params: lsp.DidOpenTextDocumentParams):
    """Parse each document when it is opened"""
    logging.debug("Received %s: %s", lsp.TEXT_DOCUMENT_DID_OPEN, params)
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.compute_diagnostics(doc)


@server.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: TclspServer, params: lsp.DidOpenTextDocumentParams):
    """Parse each document when it is changed"""
    logging.debug("Received %s: %s", lsp.TEXT_DOCUMENT_DID_CHANGE, params)
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.compute_diagnostics(doc)


@server.feature(
    lsp.TEXT_DOCUMENT_DIAGNOSTIC,
    lsp.DiagnosticOptions(
        identifier="pull-diagnostics",
        inter_file_dependencies=False,
        # We could support workspace diagnostics, although an implementation based on
        # the pygls tutorial seems to add client-server noise for no benefit (it ends up
        # replying to a frequent workspace diagnostics request with "unchanged"
        # messages).
        workspace_diagnostics=False,
    ),
)
def document_diagnostic(ls: TclspServer, params: lsp.DocumentDiagnosticParams):
    """Return diagnostics for the requested document"""
    logging.debug("Received %s: %s", lsp.TEXT_DOCUMENT_DIAGNOSTIC, params)

    was_cached = True
    if (uri := params.text_document.uri) not in ls.diagnostics:
        was_cached = False
        doc = ls.workspace.get_text_document(uri)
        ls.compute_diagnostics(doc)

    version, diagnostics = ls.diagnostics[uri]
    result_id = f"{uri}@{version}"

    if was_cached and result_id == params.previous_result_id:
        return lsp.UnchangedDocumentDiagnosticReport(result_id)

    return lsp.FullDocumentDiagnosticReport(items=diagnostics, result_id=result_id)


@server.feature(lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES)
def change_watched_files(ls: TclspServer, params: lsp.DidChangeWatchedFilesParams):
    logging.debug("Received %s: %s", lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES, params)

    # Clear diagnostics cache so they get recalculated when requested
    ls.diagnostics = {}

    ls.load_configs()
    if ls.client_supports_refresh:
        ls.lsp.send_request(lsp.WORKSPACE_DIAGNOSTIC_REFRESH, None)


@server.feature(lsp.TEXT_DOCUMENT_FORMATTING)
def format_document(ls: TclspServer, params: lsp.DocumentFormattingParams):
    """Format the entire document"""
    doc = ls.workspace.get_text_document(params.text_document.uri)

    source = doc.source
    start = lsp.Position(line=0, character=0)
    last_line = source.rsplit("\n", 1)[-1]
    end = lsp.Position(line=source.count("\n"), character=len(last_line))

    formatted = ls.format(doc, params.options)
    return [
        lsp.TextEdit(
            range=lsp.Range(start=start, end=end),
            new_text=formatted,
        )
    ]


@server.feature(lsp.TEXT_DOCUMENT_RANGE_FORMATTING)
def format_range(ls: TclspServer, params: lsp.DocumentRangeFormattingParams):
    """Format the given range with a document"""
    doc = ls.workspace.get_text_document(params.text_document.uri)

    # Round up range to full lines.
    start_line = params.range.start.line
    end_line = params.range.end.line
    if params.range.end.character > 0:
        end_line += 1
    range = lsp.Range(
        start=lsp.Position(line=start_line, character=0),
        end=lsp.Position(line=end_line, character=0),
    )

    start = doc.offset_at_position(range.start)
    end = doc.offset_at_position(range.end)

    try:
        formatted = ls.format(doc, params.options, range=(start, end))
    except TclSyntaxError:
        return None

    return [
        lsp.TextEdit(
            range=range,
            new_text=formatted,
        )
    ]


@server.feature(lsp.INITIALIZE)
def initialize(ls: TclspServer, params: lsp.InitializeParams) -> None:
    if params.initialization_options is None:
        return

    # Apply settings provided on initialization. The schema was copied from the template
    # that the tclint-vscode extension is based on.
    globalSettings = params.initialization_options.get("globalSettings", {})
    if globalSettings.get("configPath"):
        path = Path(globalSettings["configPath"]).expanduser()
        if not path.is_absolute():
            ls.show_message(
                f"Warning: expected global config path to be absolute, got {path}"
            )
        else:
            ls.global_settings.config_file = path

    for settings in params.initialization_options.get("settings", []):
        root = Path(settings["cwd"])
        if root not in ls.workspace_settings:
            ls.workspace_settings[root] = ExtensionSettings()
        if settings.get("configPath"):
            path = Path(settings["configPath"]).expanduser()
            if not path.is_absolute():
                path = root / path
            ls.workspace_settings[root].config_file = path


@server.feature(lsp.INITIALIZED)
def init(ls: TclspServer, params: lsp.InitializeParams):
    """Registers file watchers on config filenames so that we can reload configs and
    refresh diagnostics if they've changed.

    Based on code snippet in
    https://github.com/openlawlibrary/pygls/issues/376#issuecomment-1717656614.
    """
    capabilities = ls.client_capabilities.workspace

    try:
        ls.client_supports_refresh = (
            capabilities.diagnostics.refresh_support  # type: ignore[union-attr]
        )
    except AttributeError:
        ls.client_supports_refresh = False

    try:
        client_supports_watched_files_registration = (
            capabilities.did_change_watched_files.dynamic_registration  # type: ignore[union-attr] # noqa: E501
        )
    except AttributeError:
        client_supports_watched_files_registration = False

    if client_supports_watched_files_registration:
        watchers = []
        for filename in (*DEFAULT_CONFIGS, "pyproject.toml"):
            pattern = f"**/{filename}"
            watchers.append(lsp.FileSystemWatcher(glob_pattern=pattern))

        for settings in (ls.global_settings, *ls.workspace_settings.values()):
            if settings.config_file is not None:
                watchers.append(
                    lsp.FileSystemWatcher(glob_pattern=settings.config_file)
                )

        ls.register_capability(
            lsp.RegistrationParams(
                registrations=[
                    lsp.Registration(
                        id=str(uuid.uuid4()),
                        method=lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES,
                        register_options=lsp.DidChangeWatchedFilesRegistrationOptions(
                            watchers=watchers
                        ),
                    )
                ]
            )
        )

    ls.load_configs()


def main():
    parser = argparse.ArgumentParser("tclsp")
    log_levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }
    parser.add_argument(
        "-l",
        "--log-level",
        default="info",
        type=lambda x: x.lower(),
        help="set the log level. defaults to info",
        choices=log_levels.keys(),
    )
    args = parser.parse_args()
    logging.basicConfig(level=log_levels[args.log_level], format="%(message)s")

    server.start_io()


if __name__ == "__main__":
    main()



"""Parse-time handling of Tcl's builtin commands.

Based on Tcl 8.6, https://www.tcl-lang.org/man/tcl8.6/TclCmd/contents.htm.

Note that the following commands are not currently supported. If support for any of
these would be helpful for your use case, please file an issue.

- Anything related to TclOO:
  - https://www.tcl.tk/man/tcl/TclCmd/my.html
  - https://www.tcl.tk/man/tcl/TclCmd/next.html
  - https://www.tcl.tk/man/tcl/TclCmd/class.html
  - https://www.tcl.tk/man/tcl/TclCmd/copy.html
  - https://www.tcl.tk/man/tcl/TclCmd/define.html
  - https://www.tcl.tk/man/tcl/TclCmd/object.html
  - https://www.tcl.tk/man/tcl/TclCmd/self.html

- Things that are imported via `package require`
  - https://www.tcl.tk/man/tcl/TclCmd/dde.html
  - https://www.tcl.tk/man/tcl/TclCmd/http.html
  - https://www.tcl.tk/man/tcl/TclCmd/msgcat.html
  - https://www.tcl.tk/man/tcl/TclCmd/platform.html
  - https://www.tcl.tk/man/tcl/TclCmd/platform_shell.html
  - https://www.tcl.tk/man/tcl/TclCmd/transchan.html
  - https://www.tcl.tk/man/tcl/TclCmd/tcltest.html

- Tcl library commands: https://www.tcl.tk/man/tcl/TclCmd/library.html

- The "unknown" command: https://www.tcl.tk/man/tcl/TclCmd/unknown.html

- Math ops:
  - https://www.tcl.tk/man/tcl/TclCmd/mathfunc.html
  - https://www.tcl.tk/man/tcl/TclCmd/mathop.html
"""

from tclint.commands.checks import (
    CommandArgError,
    check_count,
    eval,
)
from tclint.commands.schema import commands_schema
from tclint.syntax_tree import BareWord


def _check_code(arg):
    """Check 'code' argument used by return and try."""

    val = arg.contents
    if val is None:
        return

    try:
        int(val)
    except ValueError:
        pass
    else:
        return

    if val in {"ok", "error", "return", "break", "continue"}:
        return

    raise CommandArgError(
        f"got {val}, expected one of ok, error, return, break, continue, or an integer"
    )


def _after(args, parser):
    """after ms [script...]"""
    # ref: https://www.tcl.tk/man/tcl/TclCmd/after.html

    script_arg = []
    if len(args) > 1:
        script_arg = eval(args[1:], parser, "after")

    return args[0:1] + script_arg


def _after_cancel(args, parser):
    """after id|(script...)"""
    # ref: https://www.tcl.tk/man/tcl/TclCmd/after.html
    check_count("after cancel", 1, None)

    # TODO: raise warning about not checking code

    return None


def _after_idle(args, parser):
    """after idle [script...]"""
    # ref: https://www.tcl.tk/man/tcl/TclCmd/after.html
    return eval(args, parser, "after idle")


def _apply(args, parser):
    """apply func [arg...]"""
    # ref: https://www.tcl.tk/man/tcl/TclCmd/apply.html
    if len(args) < 1:
        raise CommandArgError(
            f"not enough args to apply: got {len(args)}, expected at least 1"
        )

    func_list = parser.parse_list(args[0])
    list_len = len(func_list.children)
    if list_len < 2 or list_len > 3:
        raise CommandArgError(
            f"Invalid first argument to apply: got list of {list_len} elements,"
            " expected 2 or 3"
        )

    body = parser.parse_script(func_list.children[1])
    func_list.children[1] = body

    return [func_list] + args[1:]


_array = {
    "subcommands": {
        "anymore": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "searchId", "value": {"type": "any"}, "required": True},
            ]
        },
        "donesearch": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "searchId", "value": {"type": "any"}, "required": True},
            ]
        },
        "exists": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
            ]
        },
        "get": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "pattern", "value": {"type": "any"}, "required": False},
            ]
        },
        "names": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "mode", "value": {"type": "any"}, "required": False},
                {"name": "pattern", "value": {"type": "any"}, "required": False},
            ]
        },
        "nextelement": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "searchId", "value": {"type": "any"}, "required": True},
            ]
        },
        "set": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "list", "value": {"type": "any"}, "required": True},
            ]
        },
        "size": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
            ]
        },
        "startsearch": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
            ]
        },
        "statistics": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
            ]
        },
        "unset": {
            "positionals": [
                {"name": "arrayName", "value": {"type": "any"}, "required": True},
                {"name": "pattern", "value": {"type": "any"}, "required": False},
            ]
        },
    },
}


def _catch(args, parser):
    """catch script [resultVarName] [optionsVarName]"""
    if len(args) < 1:
        raise CommandArgError(
            f"not enough args to catch: got {len(args)}, expected at least 1"
        )
    if len(args) > 3:
        raise CommandArgError(
            f"too many args to catch: got {len(args)}, expected no more than 3"
        )

    return [parser.parse_script(args[0])] + args[1:]


_chan = {
    "subcommands": {
        "blocked": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
            ]
        },
        "close": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "direction", "value": {"type": "any"}, "required": False},
            ]
        },
        "configure": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "options", "value": {"type": "variadic"}, "required": False},
            ],
        },
        "copy": {
            "positionals": [
                {"name": "inputChan", "value": {"type": "any"}, "required": True},
                {"name": "outputChan", "value": {"type": "any"}, "required": True},
                {"name": "options", "value": {"type": "variadic"}, "required": False},
            ],
        },
        "create": {
            "positionals": [
                {"name": "mode", "value": {"type": "any"}, "required": True},
                {"name": "cmdPrefix", "value": {"type": "any"}, "required": True},
            ]
        },
        "eof": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
            ]
        },
        "event": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "event", "value": {"type": "any"}, "required": True},
                # TODO: parse this as script
                {"name": "script", "value": {"type": "any"}, "required": False},
            ]
        },
        "flush": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
            ]
        },
        "gets": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "varName", "value": {"type": "any"}, "required": False},
            ]
        },
        "names": {
            "positionals": [
                {"name": "pattern", "value": {"type": "any"}, "required": False},
            ]
        },
        "pending": {
            "positionals": [
                {"name": "mode", "value": {"type": "any"}, "required": True},
                {"name": "channelId", "value": {"type": "any"}, "required": True},
            ]
        },
        "pipe": {},
        "pop": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
            ]
        },
        "postevent": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "eventSpec", "value": {"type": "any"}, "required": True},
            ]
        },
        "push": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "cmdPrefix", "value": {"type": "any"}, "required": True},
            ]
        },
        "puts": {
            "positionals": [
                {"name": "-nonewline", "value": {"type": "any"}, "required": False},
                {"name": "channelId", "value": {"type": "any"}, "required": False},
                {"name": "string", "value": {"type": "any"}, "required": True},
            ],
        },
        "read": {
            "positionals": [
                {"name": "-nonewline", "value": {"type": "any"}, "required": False},
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "numChars", "value": {"type": "any"}, "required": False},
            ],
        },
        "seek": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "offset", "value": {"type": "any"}, "required": True},
                {"name": "origin", "value": {"type": "any"}, "required": False},
            ]
        },
        "tell": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
            ]
        },
        "truncate": {
            "positionals": [
                {"name": "channelId", "value": {"type": "any"}, "required": True},
                {"name": "length", "value": {"type": "any"}, "required": False},
            ]
        },
    },
}


def _dict_filter(args, parser):
    """dict filter <dictionaryValue> <filterType> <arg> [arg...]
    dict filter <dictionaryValue> key [globPattern...]
    dict filter <dictionaryValue> value [globPattern...]
    """

    # ref: https://www.tcl.tk/man/tcl/TclCmd/dict.html#M8

    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'dict filter': got {len(args)}, expected at least 2"
        )

    if args[1].contents not in {"key", "script", "value"}:
        raise CommandArgError(
            "invalid argument to 'dict filter': expected filter type to be one of key,"
            " script, or value"
        )

    if args[1].contents == "script":
        kv_pair = parser.parse_list(args[2])
        list_len = len(kv_pair.children)
        if len(kv_pair.children) != 2:
            raise CommandArgError(
                "invalid argument to 'dict filter': expected list of 2 elements in"
                f" second-to-last argument, got {list_len}"
            )
        return args[0:2] + [kv_pair, parser.parse_script(args[3])]

    return None


def _dict_map_for(cmd):
    def check(args, parser):
        if len(args) != 3:
            raise CommandArgError(
                f"wrong # of args to '{cmd}': got {len(args)}, expected 3"
            )

        # TODO: might be worth checking that arg[0] is a pair?

        return args[0:2] + [parser.parse_script(args[2])]

    return check


def _dict_update(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/dict.html#M25

    if len(args) < 4:
        raise CommandArgError(
            f"not enough args to 'dict update': got {len(args)}, expected at least 4"
        )

    if len(args) % 2 != 0:
        raise CommandArgError(
            "invalid # of args to 'dict update': expected an even number"
        )

    return args[0:-1] + [parser.parse_script(args[-1])]


def _dict_with(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/dict.html#M27

    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'dict with': got {len(args)}, expected at least 2"
        )

    return args[0:-1] + [parser.parse_script(args[-1])]


def _eval(args, parser):
    return eval(args, parser, "eval")


def _expr(args, parser):
    if len(args) == 0:
        raise CommandArgError("not enough args to 'expr': got 0, expected at least 1")

    # Handle single argument consisting of BareWord, BracedWord, or concrete QuotedWord.
    if len(args) == 1 and args[0].contents is not None:
        # this method will handle the `node.contents is None` case fine, but
        # will throw an error. We'll instead pass thru silently, since that error
        # will be caught by a separate lint check.
        return [parser.parse_expression(args[0])]

    # Handle multiple BareWord arguments. Non-BareWords are hard to handle in this case,
    # since we need to pop the contents out of quoted or braced words, but then we have
    # no way of storing the original info about these words in the syntax tree.
    contents = ""
    last_pos = args[0].pos
    for arg in args:
        if not isinstance(arg, BareWord):
            return None

        if arg.pos[0] != last_pos[0]:
            contents += "\n" * (arg.pos[0] - last_pos[0])
            contents += " " * (arg.pos[1] - 1)
        else:
            contents += " " * (arg.pos[1] - last_pos[1])
        contents += arg.contents
        last_pos = arg.end_pos

    node = BareWord(contents, pos=args[0].pos, end_pos=args[-1].end_pos)
    return [parser.parse_expression(node)]


def _fileevent(args, parser):
    # ref: https://www.tcl.tk/man/tcl8.4/TclCmd/fileevent.html
    # TODO: implement
    raise CommandArgError(
        "argument parsing for 'fileevent' not implemented, script argument will not be"
        " checked for violations"
    )


def _for(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/for.html
    if len(args) != 4:
        raise CommandArgError(f"wrong # of args to for: got {len(args)}, expected 4")

    return [
        parser.parse_script(args[0]),
        parser.parse_expression(args[1]),
        parser.parse_script(args[2]),
        parser.parse_script(args[3]),
    ]


def _foreach(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/foreach.html
    if len(args) < 3:
        raise CommandArgError(
            f"insufficient args to foreach: got {len(args)}, expected at least 3"
        )

    # last argument is script body
    return args[0:-1] + [parser.parse_script(args[-1])]


def _if(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/if.html
    # TODO: make arg checking strict

    new_args = []

    new_args.append(parser.parse_expression(args[0]))

    while len(new_args) < len(args):
        arg = args[len(new_args)]

        if arg.contents == "then" or arg.contents == "else":
            new_args.append(arg)
            continue
        if arg.contents == "elseif":
            new_args.append(arg)
            new_args.append(parser.parse_expression(args[len(new_args)]))
            continue

        arg = parser.parse_script(arg)
        new_args.append(arg)

    return new_args


def _interp_eval(args, parser):
    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'interp eval': got {len(args)}, expected at least 2"
        )
    return args[0:1] + eval(args[1:], parser, "interp eval")


def _lmap(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/lmap.html
    if len(args) < 3:
        raise CommandArgError(
            f"not enough args to lmap: got {len(args)}, expected at least 3"
        )

    return args[:-1] + [parser.parse_script(args[-1])]


def _namespace_code(args, parser):
    # ref: https://www.tcl.tk/man/tcl8.4/TclCmd/namespace.html#M6
    # TODO: seems like a possible pattern is to execute things in these scripts
    # with additional args provided, so command-args checks within this might
    # actually be false positive. will keep as-is for now though.
    return [parser.parse_script(args[0])]


def _namespace_eval(args, parser):
    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'namespace eval': got {len(args)}, expected at least 2"
        )
    return args[0:1] + eval(args[1:], parser, "namespace eval")


def _namespace_inscope(args, parser):
    # ref: https://www.tcl.tk/man/tcl8.4/TclCmd/namespace.html#M14
    raise CommandArgError(
        "'namespace inscope' is not meant to be called directly, consider using"
        " 'namespace code' or 'namespace eval' instead"
    )


def _package_ifneeded(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/package.html

    # TODO: implement

    # one issue with this one - it seems like calls to package ifneeded are
    # often generated by pkg_MkIndex and these calls won't lint clean. Probably
    # need a special case to ensure that these don't generate violations

    raise CommandArgError(
        "argument parsing for 'package ifneeded' not implemented, any script argument"
        " will not be checked for violations"
    )


def _proc(args, parser):
    if len(args) != 3:
        raise CommandArgError(f"wrong # of args to proc: got {len(args)}, expected 3")

    # Parse args as list, then iterate over each item to parse arg specifier lists and
    # do some validation. We don't store non-defaulted arguments as Lists so that they
    # don't get formatted inside braces.
    arg_list = parser.parse_list(args[1])
    for i, arg in enumerate(arg_list.children):
        if isinstance(arg, BareWord):
            continue

        arg_specifier = parser.parse_list(arg)
        arg_specifier_len = len(arg_specifier.children)

        if arg_specifier_len == 2:
            arg_list.children[i] = arg_specifier
        elif arg_specifier_len != 1:
            raise CommandArgError(
                f"too many fields in argument specifier: got {arg_specifier_len},"
                " expected no more than 2"
            )

    return args[0:1] + [arg_list, parser.parse_script(args[2])]


def _return(args, parser):
    args = list(args)
    while len(args) > 0:
        option = args.pop(0).contents

        try:
            if option == "-code":
                arg = args.pop(0)
                try:
                    _check_code(arg)
                except CommandArgError as e:
                    raise CommandArgError(f"invalid value for return -code: {e}")
            elif option == "-level":
                val = args.pop(0).contents

                if val is None:
                    continue

                try:
                    if int(val) >= 0:
                        continue
                except ValueError:
                    pass

                raise CommandArgError(
                    f"invalid value for return -level: got {val}, expected a"
                    " non-negative integer"
                )
            elif option in {"-errorcode", "-errorinfo", "-errorstack", "-options"}:
                args.pop(0)
            else:
                break
        except IndexError:
            raise CommandArgError(
                f"insufficient args to return: expected value after {option}"
            )

    if len(args) > 0:
        raise CommandArgError(
            "too many arguments to return: expected no more than 1 argument after"
            " explicit options. Provide -options argument if you intend to specify"
            " additional return options."
        )

    return None


def _switch(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/switch.html
    # This one's complicated...

    # TODO: better checking of malformed switch command

    arg_contents = [arg.contents for arg in args]
    arg_i = 0

    try:
        arg_i = arg_contents.index("--") + 1
    except ValueError:
        while True:
            contents = args[arg_i].contents
            if contents in {"-exact", "-glob", "-regexp", "-nocase"}:
                arg_i += 1
            elif contents in {"-matchvar", "-indexvar"}:
                arg_i += 2
            else:
                break

    # accounts for string to be matched
    arg_i += 1

    new_args = args[0:arg_i]

    # one argument left => form where patterns and bodies are in list
    last_arg_is_list = arg_i == len(args) - 1

    if last_arg_is_list:
        pattern_and_commands_list = parser.parse_list(args[arg_i])
        new_args.append(pattern_and_commands_list)
        pattern_and_commands = pattern_and_commands_list.children
    else:
        pattern_and_commands = args[arg_i:]

    if len(pattern_and_commands) % 2 != 0:
        raise CommandArgError("Expected even number of patterns and commands")

    parsed_patterns_and_commands = []
    for i, node in enumerate(pattern_and_commands):
        if i % 2 == 0:
            parsed_patterns_and_commands.append(node)
        else:
            parsed_patterns_and_commands.append(parser.parse_script(node))

    if last_arg_is_list:
        pattern_and_commands_list.children = parsed_patterns_and_commands
    else:
        new_args.extend(parsed_patterns_and_commands)

    return new_args


def _time(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/time.html
    if len(args) < 1:
        raise CommandArgError(
            f"not enough args to time: got {len(args)}, expected at least 1"
        )

    if len(args) > 2:
        raise CommandArgError(
            f"too many args to time: got {len(args)}, expected no more than 2"
        )

    if len(args) == 2:
        time = args[1].contents
        if time is not None:
            try:
                int(time)
            except ValueError:
                raise CommandArgError(
                    "invalid argument to time: expected integer for last argument"
                )

    return [parser.parse_script(args[0])] + args[1:]


def _timerate(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/timerate.html
    # timerate doesn't seem to be implemented in tclsh 8.6 for me - why?

    args = list(args)
    new_args = []

    while True:
        try:
            arg = args.pop(0)
        except IndexError:
            raise CommandArgError("invalid arguments to timerate: expected script body")

        if arg.contents in {"-direct", "-calibrate"}:
            new_args.append(arg)
        elif arg.contents in {"-overhead"}:
            new_args.append(arg)
            try:
                val = args.pop(0)
                if val.contents is not None:
                    float(val.contents)
            except (ValueError, IndexError, TypeError):
                raise CommandArgError(
                    "invalid argument to timerate: -overhead must be followed by a"
                    " double"
                )
            new_args.append(val)
        else:
            break

    new_args.append(parser.parse_script(arg))

    if len(args) > 2:
        raise CommandArgError(
            "too many arguments to timerate: expected no more than 2 arguments"
            " following script body"
        )

    try:
        [int(arg.contents) for arg in args]
    except ValueError:
        raise CommandArgError(
            "invalid argument to timerate: expected one or two integers following"
            " script body"
        )

    return new_args + args


def _try(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/try.html
    args = list(args)
    new_args = []

    while True:
        try:
            arg = args.pop(0)
        except IndexError:
            raise CommandArgError("invalid arguments to try: missing script body")
        new_args.append(parser.parse_script(arg))

        try:
            arg = args.pop(0)
        except IndexError:
            break

        new_args.append(arg)

        if arg.contents == "on":
            try:
                code = args.pop(0)
                try:
                    _check_code(code)
                except CommandArgError as e:
                    raise CommandArgError(
                        f"invalid code argument to 'on' handler in try: {e}"
                    )
                new_args.append(code)
                new_args.append(args.pop(0))
            except IndexError:
                raise CommandArgError(
                    "invalid arguments to try: expected 3 arguments after 'on' handler"
                )
        elif arg.contents == "trap":
            try:
                new_args.append(args.pop(0))
                new_args.append(args.pop(0))
            except IndexError:
                raise CommandArgError(
                    "invalid arguments to try: expected 3 arguments after 'trap'"
                    " handler"
                )
        elif arg.contents == "finally":
            continue
        else:
            raise CommandArgError(
                "invalid handler argument to try: expected one of 'on', 'trap', or"
                " 'finally'"
            )

    return new_args


def _while(args, parser):
    if len(args) != 2:
        raise CommandArgError(f"wrong # of args to while: got {len(args)}, expected 2")

    return [
        parser.parse_expression(args[0]),
        parser.parse_script(args[1]),
    ]


commands = commands_schema({
    "after": {
        "subcommands": {
            "cancel": _after_cancel,
            "idle": _after_idle,
            "info": {
                "positionals": [
                    {"name": "id", "value": {"type": "any"}, "required": False}
                ]
            },
            "": _after,
        },
    },
    "append": {
        "positionals": [
            {"name": "varname", "value": {"type": "any"}, "required": True},
            {"name": "value", "value": {"type": "variadic"}, "required": False},
        ]
    },
    "apply": _apply,
    "array": _array,
    "binary": {
        "subcommands": {
            "decode": check_count("binary decode", 2, None),
            "encode": check_count("binary encode", 2, None),
            "format": check_count("binary format", 1, None),
            "scan": check_count("binary scan", 2, None),
        },
    },
    "break": check_count("break", 0, 0),
    "catch": _catch,
    "cd": {
        "positionals": [
            {"name": "dirName", "value": {"type": "any"}, "required": False}
        ],
    },
    "chan": _chan,
    # TODO: check subcommands
    "clock": check_count("clock"),
    "close": {
        "positionals": [
            {"name": "channelId", "value": {"type": "any"}, "required": True},
            {"name": "read|write", "value": {"type": "any"}, "required": False},
        ],
    },
    "concat": {
        "positionals": [
            {"name": "arg", "value": {"type": "variadic"}, "required": True},
        ]
    },
    "continue": {},
    "coroutine": {
        "positionals": [
            {"name": "name", "value": {"type": "any"}, "required": True},
            {"name": "command", "value": {"type": "any"}, "required": True},
            {"name": "arg", "value": {"type": "variadic"}, "required": False},
        ]
    },
    "dict": {
        "subcommands": {
            "append": check_count("dict append", 2, None),
            "create": check_count("dict create"),
            "exists": check_count("dict exists", 2, None),
            "filter": _dict_filter,
            "for": _dict_map_for("dict for"),
            "get": check_count("dict get", 1, None),
            "incr": check_count("dict incr", 2, 3),
            "info": check_count("dict info", 1, 1),
            "keys": check_count("dict keys", 1, 2),
            "lappend": check_count("dict lappend", 2, None),
            "map": _dict_map_for("dict map"),
            "merge": check_count("dict merge"),
            "remove": check_count("dict remove", 1, None),
            "replace": check_count("dict replace", 1, None),
            "set": check_count("dict set", 3, None),
            "size": check_count("dict size", 1, 1),
            "unset": check_count("dict unset", 2, None),
            "update": _dict_update,
            "values": check_count("dict values", 1, 2),
            "with": _dict_with,
        },
    },
    "encoding": {
        "subcommands": {
            "convertfrom": check_count("encoding convertfrom", 1, 2),
            "convertto": check_count("encoding convertto", 1, 2),
            "dirs": check_count("encoding dirs", 0, 1),
            "names": check_count("encoding names", 0, 0),
            "system": check_count("encoding system", 0, 1),
        },
    },
    "eof": check_count("eof", 1, 1),
    "error": check_count("error", 1, 3),
    "eval": _eval,
    "exec": check_count("exec", 1, None),
    "exit": check_count("exit", 0, 1),
    "expr": _expr,
    "fblocked": check_count("fblocked", 1, 1),
    "fconfigure": check_count("fconfigure", 1, None),
    "fcopy": check_count("fcopy", 2, 6),
    # TODO: check subcommands
    "file": check_count("file", 1, None),
    "fileevent": _fileevent,
    "flush": check_count("flush", 1, 1),
    "for": _for,
    "foreach": _foreach,
    "format": check_count("format", 1, None),
    "gets": check_count("gets", 1, 2),
    "glob": check_count("glob"),
    "global": check_count("global"),
    "history": check_count("history"),
    "if": _if,
    "incr": check_count("incr", 1, 2),
    # TODO: check subcommands
    "info": check_count("info", 1, None),
    # TODO: check other subcommands
    "interp": {
        "subcommands": {
            "eval": _interp_eval,
            "": check_count("interp", 1, None),
        },
    },
    "join": check_count("join", 1, 2),
    "lappend": check_count("lappend", 1, None),
    "lassign": check_count("lassign", 1, None),
    "lindex": check_count("lindex", 1, None),
    "linsert": check_count("linsert", 2, None),
    "list": check_count("list", 0, None),
    "llength": check_count("llength", 1, 1),
    "lrepeat": check_count("lrepeat", 1, None),
    "lreplace": check_count("lreplace", 3, None),
    "lreverse": check_count("lreverse", 1, 1),
    "lset": check_count("lset", 2, None),
    "lsort": check_count("lsort", 1, None),
    "lmap": _lmap,
    "load": check_count("load", 1, 6),
    "lrange": check_count("lrange", 3, 3),
    "lsearch": check_count("lsearch", 2, None),
    "memory": {
        "subcommands": {
            "active": check_count("memory active", 1, 1),
            "break_on_malloc": check_count("memory break_on_malloc", 1, 1),
            "info": check_count("memory info", 0, 0),
            # just on or off
            "init": check_count("memory init", 1, 1),
            "objs": check_count("memory objs", 1, 1),
            "onexit": check_count("memory onexit", 1, 1),
            "tag": check_count("memory tag", 1, 1),
            # just on or off
            "trace": check_count("memory trace", 1, 1),
            "trace_on_at_malloc": check_count("memory trace_on_at_malloc", 1, 1),
            # just on or off
            "validate": check_count("memory validate", 1, 1),
        },
    },
    "namespace": {
        "subcommands": {
            "children": check_count("namespace children", 0, 2),
            "code": _namespace_code,
            "current": check_count("namespace current", 0, 0),
            "delete": None,
            "eval": _namespace_eval,
            "exists": check_count("namespace exists", 1, 1),
            "export": None,
            "forget": None,
            "import": None,
            "inscope": _namespace_inscope,
            "origin": check_count("namespace origin", 1, 1),
            "parent": check_count("namespace parent", 0, 1),
            "qualifiers": check_count("namespace qualifiers", 1, 1),
            "tail": check_count("namespace tail", 1, 1),
            "which": check_count("namespace which", 1, 2),
            "ensemble": {
                "subcommands": {
                    "create": None,
                    "configure": check_count("namespace ensemble configure", 1, None),
                    "exists": check_count("namespace ensemble exists", 1, 1),
                },
            },
        },
    },
    "open": check_count("open", 1, 3),
    "package": {
        "subcommands": {
            "forget": None,
            "ifneeded": _package_ifneeded,
            "names": check_count("package names", 0, 0),
            "present": check_count("package present", 0, None),
            "provide": check_count("package provide", 1, 2),
            "require": check_count("package require", 1, None),
            "unknown": check_count("package unknown", 1, None),
            "vcompare": check_count("package vcompare", 2, 2),
            "versions": check_count("package versions", 1, 1),
            "vsatisfies": check_count("package vsatisfies", 2, None),
            "prefer": check_count("package prefer", 1, 1),
        },
    },
    "pid": check_count("pid", 0, 1),
    "pkg::create": check_count("pkg::create", 2, None),
    "pkg_mkIndex": check_count("pkg_mkIndex", 1, None),
    "proc": _proc,
    "puts": {
        "positionals": [
            {"name": "-nonewline", "value": {"type": "any"}, "required": False},
            {"name": "channelId", "value": {"type": "any"}, "required": False},
            {"name": "string", "value": {"type": "any"}, "required": True},
        ],
    },
    "pwd": check_count("pwd", 0, 0),
    "read": check_count("read", 1, 2),
    "regexp": check_count("regexp", 2, None),
    "regsub": check_count("regsub", 3, None),
    "rename": check_count("rename", 2, 2),
    "return": _return,
    # TODO: check subcommands
    "safe": check_count("safe", 1, None),
    "scan": check_count("scan", 2, None),
    "seek": check_count("seek", 2, 3),
    "set": check_count("set", 1, 2),
    "socket": check_count("socket", 2, None),
    "source": check_count("source", 1, 3),
    "split": check_count("split", 1, 2),
    # TODO: check subcommands
    "string": check_count("string", 2, None),
    "subst": check_count("subst", 1, 4),
    "switch": _switch,
    "tailcall": check_count("tailcall", 1, None),
    "tcl::prefix": {
        "subcommands": {
            "all": check_count("tcl::prefix all", 2, 2),
            "longest": check_count("tcl::prefix longest", 2, 2),
            "match": check_count("tcl::prefix match", 2, None),
        },
    },
    "tell": check_count("tell", 1, 1),
    "throw": check_count("throw", 2, 2),
    "time": _time,
    "timerate": _timerate,
    "tcl::tm::path": {
        "subcommands": {
            "add": check_count("tcl::tm::path add"),
            "remove": check_count("tcl::tm::path remove"),
            "list": check_count("tcl::tm::path list", 0, 0),
        },
    },
    "tcl::tm::roots": check_count("tcl::tm::roots"),
    # TODO: check subcommands
    "trace": check_count("trace", 2, None),
    "try": _try,
    "unload": check_count("unload", 1, 6),
    "unset": check_count("unset"),
    "update": check_count("update", 0, 1),
    "uplevel": check_count("uplevel", 1, None),
    "upvar": check_count("upvar", 2, None),
    "variable": check_count("variable", 1, None),
    "vwait": check_count("vwait", 1, 1),
    "while": _while,
    "yield": {
        "positionals": [
            {"name": "value", "value": {"type": "any"}, "required": False},
        ]
    },
    "yieldto": {
        "positionals": [
            {"name": "command", "value": {"type": "any"}, "required": True},
            {"name": "arg", "value": {"type": "variadic"}, "required": False},
        ]
    },
    # TODO: check subcommands
    "zlib": check_count("zlib", 3, None),
})



from importlib_metadata import entry_points
import json
import pathlib
from typing import Dict, Optional
from types import ModuleType

import voluptuous

from tclint.commands.schema import schema as command_schema


class _PluginManager:
    def __init__(self):
        self._loaded = {}
        self._installed = {}
        self._loaded_specs = {}
        for plugin in entry_points(group="tclint.plugins"):
            if plugin.name in self._installed:
                print(f"Warning: found duplicate definitions for plugin {plugin.name}")
            self._installed[plugin.name] = plugin

    def load(self, name: str) -> Optional[Dict]:
        if name in self._loaded:
            return self._loaded[name]

        mod = self._load(name)
        self._loaded[name] = mod
        return mod

    def load_from_spec(self, path: pathlib.Path) -> Optional[Dict]:
        if path in self._loaded_specs:
            return self._loaded_specs[path]

        spec = self._load_from_spec(path)
        self._loaded_specs[path] = spec
        return spec

    def _load_from_spec(self, path: pathlib.Path) -> Optional[Dict]:
        try:
            with open(path.expanduser(), "r") as f:
                spec = json.load(f)
        except (FileNotFoundError, RuntimeError):
            print(f"Warning: command spec {path} not found, skipping...")
            return None

        try:
            # Apply defaults and validate the spec.
            spec = command_schema(spec)
        except voluptuous.Invalid as e:
            print(f"Warning: invalid command spec {path}: {e}")
            return None

        return spec["commands"]

    def get_mod(self, name: str) -> Optional[ModuleType]:
        if name not in self._installed:
            print(f"Warning: plugin {name} is not installed")
            return None

        plugin = self._installed[name]

        try:
            module = plugin.load()
        except Exception as e:
            print(f"Warning: error loading plugin {name}: {e}")
            return None

        return module

    def _load(self, name: str):
        module = self.get_mod(name)
        if module is None:
            print(f"Skipping requested plugin {name}")
            return None

        if not hasattr(module, "commands"):
            print(f"Warning: skipping plugin {name} since it does not define commands")
            return None

        return getattr(module, "commands")


# TODO: we'll probably want to construct this in the tclint entry point and pass
# it around rather than using a singleton instance, but this made for an easier
# refactor.
PluginManager = _PluginManager()



import pytest

from voluptuous import Invalid

from tclint.commands.schema import schema


def test_valid_command_spec():
    command_spec = {
        "name": "test_plugin",
        "commands": {
            "command1": {
                "positionals": [
                    {"name": "arg1", "required": True, "value": {"type": "any"}},
                    {"name": "arg2", "required": False, "value": {"type": "variadic"}},
                ],
                "switches": {
                    "-switch1": {
                        "required": False,
                        "value": {"type": "any"},
                        "repeated": False,
                        "metavar": "value1",
                    },
                    "-switch2": {
                        "required": True,
                        "value": None,
                        "repeated": True,
                    },
                },
            }
        },
    }

    schema(command_spec)


def test_missing_required_field():
    invalid_command_spec = {
        "name": "test_plugin",
        "commands": {
            "command1": {
                "switches": {
                    # missing "required"
                    "-switch1": {"value": {"type": "any"}, "repeated": False},
                },
            }
        },
    }

    with pytest.raises(Invalid) as excinfo:
        schema(invalid_command_spec)
    print("test_missing_required_field:", excinfo.value)


def test_invalid_datatype():
    invalid_minmax_spec = {
        "name": "test_plugin",
        "commands": {
            "command1": {
                "switches": {
                    "-switch1": {
                        "required": True,
                        "value": {"type": "any"},
                        "repeated": "invalid",  # Should be a boolean
                    }
                },
            }
        },
    }

    with pytest.raises(Invalid) as excinfo:
        schema(invalid_minmax_spec)
    print("test_invalid_minmax_values:", excinfo.value)


def test_extra_property():
    invalid_spec = {
        "name": "test_plugin",
        "commands": {
            "command1": {
                "positionals": [
                    {
                        "name": "arg1",
                        "required": True,
                        "value": {"type": "any"},
                        "extra_field": "not allowed",  # Extra property not in schema
                    }
                ],
            }
        },
    }

    with pytest.raises(Invalid) as excinfo:
        schema(invalid_spec)
    print("test_extra_property:", excinfo.value)


def test_command_no_validation():
    command_spec = {
        "name": "test_plugin",
        "commands": {
            "command": None,
        },
    }

    schema(command_spec)


def test_subcommands():
    command_spec = {
        "name": "test_plugin",
        "commands": {
            "command1": {
                "subcommands": {
                    "subcommand1": {
                        "positionals": [
                            {
                                "name": "arg1",
                                "required": True,
                                "value": {"type": "any"},
                            },
                        ],
                    },
                    "subcommand2": None,
                    "": {
                        "switches": {
                            "-switch1": {
                                "required": False,
                                "value": {"type": "any"},
                                "repeated": False,
                            },
                        },
                    },
                }
            }
        },
    }
    schema(command_spec)


def test_applies_defaults():
    command_spec = {
        "name": "test_plugin",
        "commands": {"command1": {}, "command2": {"subcommands": {"subcommand1": {}}}},
    }
    validated = schema(command_spec)
    for key in ("positionals", "switches"):
        assert key in validated["commands"]["command1"]
        assert key in validated["commands"]["command2"]["subcommands"]["subcommand1"]


def test_functions():
    """For Python plugins, command specs can be functions."""
    command_spec = {
        "name": "test_plugin",
        "commands": {
            "command1": lambda args, parser: None,
            "command2": {"subcommands": {"subcommand1": lambda args, parser: None}},
        },
    }
    schema(command_spec)



from pathlib import Path
import shutil
import sys

from lsprotocol import types as lsp
import pytest
import pytest_lsp

MY_DIR = Path(__file__).parent.resolve()
LSP_BIN = MY_DIR / ".." / "src" / "tclint" / "cli" / "tclsp.py"


def get_capabilities(client: str) -> lsp.ClientCapabilities:
    capabilities = pytest_lsp.client_capabilities(client)
    # Need to nullify these capabilities, since pytest_lsp doesn't support them
    try:
        capabilities.workspace.diagnostics.refresh_support = False
    except AttributeError:
        pass
    try:
        capabilities.workspace.did_change_watched_files.dynamic_registration = False
    except AttributeError:
        pass
    return capabilities


@pytest_lsp.fixture(
    config=pytest_lsp.ClientServerConfig(
        # -I ensures the server imports the tclint package instead of the entry point.
        server_command=[sys.executable, "-I", str(LSP_BIN)]
    )
)
async def client(lsp_client: pytest_lsp.LanguageClient, request):
    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_diagnostics(client: pytest_lsp.LanguageClient):
    """Basic diagnostics test."""
    params = lsp.InitializeParams(capabilities=get_capabilities("visual-studio-code"))
    await client.initialize_session(params)

    document = MY_DIR / "data" / "dirty.tcl"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=f"file://{document}")
        )
    )
    assert results is not None
    assert len(results.items) == 2
    items = sorted(
        results.items, key=lambda r: (r.range.start.line, r.range.start.character)
    )

    assert items[0].message == "unnecessary command substitution within expression"
    assert items[0].code == "redundant-expr"
    assert items[0].range.start.line == 5
    assert items[0].range.start.character == 10
    assert items[0].range.end.line == 5
    assert items[0].range.end.character == 23

    assert (
        items[1].message == "expression with substitutions should be enclosed by braces"
    )
    assert items[1].code == "unbraced-expr"
    assert items[1].range.start.line == 5
    assert items[1].range.start.character == 16
    assert items[1].range.end.line == 5
    assert items[1].range.end.character == 22


@pytest.mark.asyncio
async def test_format(client: pytest_lsp.LanguageClient, tmp_path):
    """Formatting test."""
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=tmp_path.as_uri(), name=tmp_path.name)
        ],
    )
    await client.initialize_session(params)

    document = tmp_path / "source.tcl"

    INDENT_TEST = """if {{1}} {{
{}puts "hello"
}}
"""
    for source, expected, end in (
        ("foo", "foo\n", (0, 3)),
        ("foo\n\n\n\n\nbar\n", "foo\n\n\nbar\n", (6, 0)),
        (INDENT_TEST.format(""), INDENT_TEST.format("   "), (3, 0)),
    ):
        with open(document, "w") as f:
            f.write(source)

        result = await client.text_document_formatting_async(
            params=lsp.DocumentFormattingParams(
                text_document=lsp.TextDocumentIdentifier(uri=document.as_uri()),
                options=lsp.FormattingOptions(tab_size=3, insert_spaces=True),
            ),
        )
        assert len(result) == 1
        result = result[0]

        assert result.new_text == expected
        assert result.range.start.line == 0
        assert result.range.start.character == 0
        assert result.range.end.line == end[0]
        assert result.range.end.character == end[1]

    # Check that tclint config overrides client config
    config = tmp_path / "tclint.toml"
    with open(config, "w") as f:
        f.write("""[style]
indent = 'tab'""")

    client.workspace_did_change_watched_files(
        # `changes` isn't used by our server impl, so we can cheat and keep this empty
        params=lsp.DidChangeWatchedFilesParams(changes=[])
    )

    result = await client.text_document_formatting_async(
        params=lsp.DocumentFormattingParams(
            text_document=lsp.TextDocumentIdentifier(uri=document.as_uri()),
            options=lsp.FormattingOptions(tab_size=3, insert_spaces=True),
        ),
    )
    assert len(result) == 1
    result = result[0]

    assert result.new_text == INDENT_TEST.format("\t")
    assert result.range.start.line == 0
    assert result.range.start.character == 0
    assert result.range.end.line == 3
    assert result.range.end.character == 0


@pytest.mark.asyncio
async def test_format_range(client: pytest_lsp.LanguageClient, tmp_path):
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=tmp_path.as_uri(), name=tmp_path.name)
        ],
    )
    await client.initialize_session(params)

    document = tmp_path / "source.tcl"

    source = r"""
puts "hello"

    if {1} {
  puts  "world"
 }

puts "goodbye"
"""

    expected = r"""
    if {1} {
       puts "world"
    }
""".lstrip("\n")

    with open(document, "w") as f:
        f.write(source)

    result = await client.text_document_range_formatting_async(
        params=lsp.DocumentRangeFormattingParams(
            text_document=lsp.TextDocumentIdentifier(uri=document.as_uri()),
            options=lsp.FormattingOptions(tab_size=3, insert_spaces=True),
            range=lsp.Range(
                start=lsp.Position(line=3, character=8),
                end=lsp.Position(line=5, character=1),
            ),
        ),
    )
    assert len(result) == 1
    result = result[0]

    assert result.new_text == expected
    assert result.range.start.line == 3
    assert result.range.start.character == 0
    assert result.range.end.line == 6
    assert result.range.end.character == 0


@pytest.mark.asyncio
async def test_config(client: pytest_lsp.LanguageClient, tmp_path_factory):
    """Tests configuration reading, specifically:
    1) That configs for multiple workspaces are found and applied correctly.
    2) That diagnostics are refreshed correctly when config is updated.

    Note proper use of the client capabilities we rely on for file watching/diagnostic
    refresh aren't captured by this test, the client is controlled manually.
    """
    # Set up workspaces
    document = MY_DIR / "data" / "dirty.tcl"

    ws_foo = tmp_path_factory.mktemp("foo")
    with open(ws_foo / "tclint.toml", "w") as f:
        f.write("ignore = ['redundant-expr']")
    doc_foo = Path(shutil.copy(document, ws_foo))

    ws_bar = tmp_path_factory.mktemp("bar")
    with open(ws_bar / "tclint.toml", "w") as f:
        f.write("ignore = ['unbraced-expr']")
    doc_bar = Path(shutil.copy(document, ws_bar))

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=ws_foo.as_uri(), name=ws_foo.name),
            lsp.WorkspaceFolder(uri=ws_bar.as_uri(), name=ws_bar.name),
        ],
    )
    await client.initialize_session(params)

    # Check diagnostics for workspace "foo"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_foo.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 1
    assert results.items[0].code == "unbraced-expr"

    # Check diagnostics for workspace "bar"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_bar.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 1
    assert results.items[0].code == "redundant-expr"

    # Make sure diagnostics refresh when config file is updated
    with open(ws_foo / "tclint.toml", "w") as f:
        f.write("ignore = ['redundant-expr', 'unbraced-expr']")
    client.workspace_did_change_watched_files(
        # `changes` isn't used by our server impl, so we can cheat and keep this empty
        params=lsp.DidChangeWatchedFilesParams(changes=[])
    )

    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_foo.as_uri())
        )
    )
    assert results.items == []


@pytest.mark.asyncio
async def test_config_extension_settings(
    client: pytest_lsp.LanguageClient, tmp_path_factory
):
    """Tests configuration reading, similar to test_config, except with paths to the
    config files supplied to the LSP via initialization params settings.
    """
    # Set up workspaces
    document = MY_DIR / "data" / "dirty.tcl"

    ws_foo = tmp_path_factory.mktemp("foo")
    config_foo = ws_foo / "nondefault.toml"
    with open(config_foo, "w") as f:
        f.write("ignore = ['redundant-expr']")
    doc_foo = Path(shutil.copy(document, ws_foo))

    ws_bar = tmp_path_factory.mktemp("bar")
    config_bar = ws_bar / "nondefault.toml"
    with open(config_bar, "w") as f:
        f.write("ignore = ['unbraced-expr']")
    doc_bar = Path(shutil.copy(document, ws_bar))

    config_global = tmp_path_factory.mktemp("global") / "nondefault.toml"
    with open(config_global, "w") as f:
        f.write("ignore = ['redundant-expr', 'unbraced-expr']")

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=ws_foo.as_uri(), name=ws_foo.name),
            lsp.WorkspaceFolder(uri=ws_bar.as_uri(), name=ws_bar.name),
        ],
        initialization_options={
            "settings": [
                {
                    "cwd": ws_foo,
                    "configPath": config_foo,
                },
                {
                    "cwd": ws_bar,
                    "configPath": config_bar,
                },
            ],
            "globalSettings": {"cwd": "/", "configPath": config_global},
        },
    )
    await client.initialize_session(params)

    # Check diagnostics for workspace "foo"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_foo.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 1
    assert results.items[0].code == "unbraced-expr"

    # Check diagnostics for workspace "bar"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_bar.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 1
    assert results.items[0].code == "redundant-expr"

    # Check diagnostics for an unpatriated file, global config should apply here
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=document.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 0


@pytest.mark.asyncio
async def test_invalid_config(client: pytest_lsp.LanguageClient, tmp_path_factory):
    # Set up workspace
    ws = tmp_path_factory.mktemp("ws")
    config = ws / "tclint.toml"
    # Invalid config
    with open(config, "w") as f:
        f.write("asdf")

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=ws.as_uri(), name=ws.name),
        ],
    )
    await client.initialize_session(params)
    await client.wait_for_notification(lsp.WINDOW_SHOW_MESSAGE)
    assert len(client.messages) == 1
    assert (
        client.messages[0].message
        == f"Error loading config file: {config}: Expected '=' after a key in a"
        " key/value pair (at end of document)"
    )


@pytest.mark.asyncio
async def test_nested_configs(client: pytest_lsp.LanguageClient, tmp_path_factory):
    # Set up workspaces
    parent_ws = tmp_path_factory.mktemp("ws")
    with open(parent_ws / "tclint.toml", "w") as f:
        f.write("ignore = []")
    child_ws = parent_ws / "child"
    child_ws.mkdir()
    with open(child_ws / "tclint.toml", "w") as f:
        f.write("ignore = ['command-args']")

    document = child_ws / "source.tcl"
    with open(document, "w") as f:
        # Missing arguments
        f.write("puts")

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=parent_ws.as_uri(), name=parent_ws.name),
            lsp.WorkspaceFolder(uri=child_ws.as_uri(), name=child_ws.name),
        ],
    )
    await client.initialize_session(params)
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=document.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 0


@pytest.mark.asyncio
async def test_lsp_exclude(client: pytest_lsp.LanguageClient, tmp_path):
    """Test that files excluded by the config are ignored."""
    document = tmp_path / "excluded.tcl"
    config = tmp_path / "tclint.toml"

    with open(document, "w") as f:
        # Missing arguments
        f.write("puts")

    with open(config, "w") as f:
        f.write(f"exclude = ['{document.name}']")

    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        initialization_options={
            "globalSettings": {
                "configPath": config,
            },
        },
    )
    await client.initialize_session(params)

    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=document.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 0



import pathlib

from tclint.format import Formatter, FormatterOpts
from tclint.parser import Parser

MY_DIR = pathlib.Path(__file__).parent.resolve()


def _test(
    script,
    expected,
    indent="  ",
    spaces_in_braces=True,
    max_blank_lines=2,
    indent_namespace_eval=True,
):
    parser = Parser()
    format = Formatter(
        FormatterOpts(
            indent=indent,
            spaces_in_braces=spaces_in_braces,
            max_blank_lines=max_blank_lines,
            indent_namespace_eval=indent_namespace_eval,
        )
    )
    out = format.format_top(script, parser)

    assert out == expected + "\n"

    original_tree = parser.parse(script)
    formatted_tree = parser.parse(out)
    assert original_tree == formatted_tree


def test_fizzbuzz():
    with open(MY_DIR / "data" / "dirty.tcl", "r") as f:
        script = f.read()

    expected = r"""
for { set i 1 } { $i < 100 } { incr i } {
  if { $i % 15 == 0 } {
    puts "FizzBuzz"
  } elseif { $i % 3 == 0 } {
    puts "Fizz"
  } elseif { [expr $i % 5] == 0 } {
    puts "Buzz"
  } else {
    puts $i
  }
}""".strip()

    _test(script, expected)


def test_blank_lines():
    """Formatter preserves up to two blank lines."""
    script = r"""
foo

foo


foo



foo"""

    expected = r"""
foo

foo


foo


foo""".strip()
    _test(script, expected)

    expected = r"""
foo

foo

foo

foo""".strip()
    _test(script, expected, max_blank_lines=1)


def test_multiple_cmds_per_line():
    """Formatter preserves commands on same line."""
    script = "foo; foo"
    _test(script, script)


def test_comments():
    """Formatter preserves comments, but will update indentation and enforce a single
    space between command and inline comment."""
    script = r"""
  # this is foo
foo;     # foo"""

    expected = r"""
# this is foo
foo ;# foo""".strip()

    _test(script, expected)


def test_switch():
    script = r"""
switch $arg {
        a {
        foo } b {
      bar
    }}
"""

    expected = r"""
switch $arg {
  a {
    foo
  } b {
    bar
  }
}""".strip()

    _test(script, expected)


def test_no_reindent_braced_word():
    """We can't add an extra level of indent in the second line of the braced word,
    since this will change the actual text."""
    script = r"""
puts \
{ one
  two }
"""

    expected = r"""
puts \
  { one
  two }""".strip()

    _test(script, expected)


def test_no_reindent_braced_word_script():
    """We can't add an extra level of indent in the second line of the braced word,
    since this will change the actual text."""
    script = r"""
if { 1 } {
puts { one
  two }
}"""

    expected = r"""
if { 1 } {
  puts { one
  two }
}""".strip()

    _test(script, expected)


def test_reindent_command_sub():
    script = r"""
puts [command \
foo]"""

    expected = r"""
puts [command \
  foo]
""".strip()

    _test(script, expected)


def test_reindent_command_sub_new_line():
    script = r"""
puts \
[command \
foo]"""

    expected = r"""
puts \
  [command \
    foo]
""".strip()

    _test(script, expected)


def test_expr_align():
    script = r"""
if {$a && $b &&
$c } { puts "asdf" }"""

    expected = r"""
if {
  $a && $b &&
  $c
} { puts "asdf" }""".strip()

    _test(script, expected)


def test_ternary_op_align():
    script = r"""
expr { $foo ? 2
    + 3 :
    4 }"""

    expected = r"""
expr {
  $foo ? 2
  + 3 :
  4
}""".strip()

    _test(script, expected)


def test_expr_command_sub_alignment():
    script = r"""
if { ![command $arg1 \
    $arg2 \
    $arg3] } {
    return true
}"""

    expected = r"""
if {
  ![command $arg1 \
    $arg2 \
    $arg3]
} {
  return true
}""".strip()

    _test(script, expected)


def test_indent_bare_word_command_sub():
    script = r"""
puts foo[bar\
    "mootown"]qwerty"""

    expected = r"""
puts foo[bar \
  "mootown"]qwerty
""".strip()

    _test(script, expected)


def test_varsub_index_preserve_newline():
    script = r"""
puts $foo(asdf \
asdf)""".strip()

    _test(script, script)


def test_varsub_indent_format():
    script = r"""
$foo(asdf$asdf[asdf \
-asdf] [qwerty \
-uiop])"""

    expected = r"""
$foo(asdf$asdf[asdf \
  -asdf] [qwerty \
  -uiop])""".strip()

    _test(script, expected)


def test_braced_varsub():
    script = r"${one_two}_three"
    _test(script, script)


def test_function():
    script = r"expr { max($a, $b) }"
    _test(script, script)


def test_dont_add_trailing_space_indent():
    script = r"""
if { 1 } {
puts "one"

puts "two"
}"""

    expected = r"""
if { 1 } {
  puts "one"

  puts "two"
}""".strip()

    _test(script, expected)


def test_paren_format():
    script = r"""
expr { 1 + (2 *
       3) }"""

    expected = r"""
expr {
  1 + (2 *
    3)
}""".strip()

    _test(script, expected)


def test_expr_alignment_nested():
    """Original indentation implementations failed when we had a doubly nested binop
    before other sub-expression types."""
    script = r"""
expr { 1 + 2 && !($foo ||
$bar) }"""

    expected = r"""
expr {
  1 + 2 && !($foo ||
    $bar)
}""".strip()

    _test(script, expected)

    script = r"""
expr { 1 + 2 + min($a,
    $b,
    $c) }"""

    expected = r"""
expr {
  1 + 2 + min($a,
    $b,
    $c)
}""".strip()

    _test(script, expected)

    script = r"""
expr { 1 + 2 * [command \
    -foo\
    -bar] }
"""

    expected = r"""
expr {
  1 + 2 * [command \
    -foo \
    -bar]
}""".strip()
    _test(script, expected)

    script = r"""
expr { $foo
* 5 + (2 * (3 * 4) + 5 + 7
* 16 + (2
* 3)) }"""

    expected = r"""
expr {
  $foo
  * 5 + (2 * (3 * 4) + 5 + 7
    * 16 + (2
      * 3))
}""".strip()

    _test(script, expected)


def test_indent_namespace_eval():
    script = r"""
namespace eval my_namespace {
    foo
}"""

    expected_no_indent = r"""
namespace eval my_namespace {
foo
}""".strip()

    expected_indent = r"""
namespace eval my_namespace {
  foo
}""".strip()

    _test(script, expected_no_indent, indent_namespace_eval=False)
    _test(script, expected_indent, indent_namespace_eval=True)


def test_indent_script_in_expr():
    script = r"""
expr {1 + 5 + [if {1} {
return 1
} else {
return 2
}]}"""

    # TODO: is this the format we want?
    expected = r"""
expr {
  1 + 5 + [if { 1 } {
    return 1
  } else {
    return 2
  }]
}""".strip()

    _test(script, expected)


def test_remove_lines_at_ends_of_script():
    script = r"""

proc foo { } {

    puts "asdf"

}

"""

    expected = r"""
proc foo { } {
  puts "asdf"
}""".strip()

    _test(script, expected)


def test_disable():
    script = r"""
command1
 command2 ; # tclfmt-disable
  command3
 command4; # tclfmt-enable
 command5"""

    expected = r"""
command1
command2 ;# tclfmt-disable
  command3
 command4; # tclfmt-enable
command5""".strip()

    _test(script, expected)


def test_quoted_expr():
    script = r"""
if "1 + 2 > 0" {
  puts "foo"
}""".strip()

    _test(script, script)


def test_multiple_commands_in_command_sub():
    script = r"""
[command; command]
""".strip()
    _test(script, script)

    script = r"""
puts [
  command1
  command2
]""".strip()
    _test(script, script)


def test_preserve_comment_line():
    script = r"""
if { 1 } { # tclint-disable-line
}""".strip()
    _test(script, script)


def test_add_indent_expr():
    script = r"""
expr {
1
}
"""

    expected = r"""
expr {
  1
}""".strip()

    _test(script, expected)


def test_empty_braces():
    script = r"if {1} {}"
    _test(script, script, spaces_in_braces=False)

    expected = r"if { 1 } { }"
    _test(script, expected, spaces_in_braces=True)


def test_function_line_breaks():
    script = r"""
expr {
  max($a,
  $b, $c)
}""".strip()

    expected = r"""
expr {
  max($a,
    $b, $c)
}""".strip()

    _test(script, expected)

    script = r"""
expr {
  max(
  $a, $b, $c)
}""".strip()

    expected = r"""
expr {
  max(
    $a, $b, $c)
}""".strip()

    _test(script, expected)

    script = r"""
expr {
  max(
    $a, $b, $c
    )
}""".strip()

    expected = r"""
expr {
  max(
    $a, $b, $c
  )
}""".strip()

    _test(script, expected)


def test_paren_line_breaks():
    script = r"""
expr {
  (
  1 + 2 + 3
  )
}""".strip()

    expected = r"""
expr {
  (
    1 + 2 + 3
  )
}""".strip()

    _test(script, expected)


def test_if_else_newline_escape_indent():
    script = r"""
if { $cond } {
  puts "true"
} else \
{
  puts "false"
}"""

    expected = r"""
if { $cond } {
  puts "true"
} else \
  {
    puts "false"
  }""".strip()

    _test(script, expected)

    script = r"""
if { $cond } {
  puts "true"
} \
else \
{
  puts "false"
}""".strip()

    expected = r"""
if { $cond } {
  puts "true"
} \
  else \
  {
    puts "false"
  }""".strip()

    _test(script, expected)


def test_multiline_unbraced_expr():
    script = r"""
expr 1 \
+ 2""".strip()

    expected = r"""
expr 1 \
  + 2""".strip()

    _test(script, expected)


def test_preserve_braces_in_unbraced_expr():
    """Test ensures that braces aren't stripped in this case. Removing them changes the
    semantic meaning of the command."""
    script = r"expr {$foo} + 2"
    _test(script, script)


def test_strip_trailing_whitespace_from_comments():
    _test("# foo ", "# foo")


def test_partial():
    script = r"""
    puts "foo"
      puts "bar"
"""
    expected = r"""
    puts "foo"
    puts "bar"
"""
    parser = Parser()
    format = Formatter(
        FormatterOpts(
            indent="    ",
            spaces_in_braces=True,
            max_blank_lines=2,
            indent_namespace_eval=True,
        )
    )

    out = format.format_partial(script, parser)

    assert out == expected

    original_tree = parser.parse(script)
    formatted_tree = parser.parse(out)
    assert original_tree == formatted_tree



from collections import defaultdict

from tclint.syntax_tree import Visitor
from tclint.violations import ALL_RULES, Rule


class CommentVisitor(Visitor):
    """Scans the tree for lint waiver comments."""

    def __init__(self):
        # line -> [rule]
        self.ignore_lines = defaultdict(set)

        self._disable_regions = {
            # rule -> line
        }

    def run(self, tree, path):
        self._path = path
        tree.accept(self, recurse=True)

        # resolve remaining disabled regions
        last_line = tree.end_pos[0]
        for rule, start_line in self._disable_regions.items():
            for line in range(start_line, last_line + 1):
                self.ignore_lines[line].add(rule)

        return self.ignore_lines

    def visit_comment(self, comment):
        contents = comment.value.strip()

        if not contents.startswith("tclint-"):
            return

        split = contents.split(" ", 1)

        command = split[0]

        rule_strs = []
        if len(split) > 1:
            rest = split[-1]
            rule_strs = rest.split("--", 1)[0]
            rule_strs = rule_strs.replace(" ", "")
            rule_strs = rule_strs.split(",")

        rules = []
        if not rule_strs:
            # default if no rules specified is all violation types
            rules = ALL_RULES
        else:
            for rule in rule_strs:
                try:
                    rules.append(Rule(rule))
                except ValueError:
                    self._warning(
                        f"unknown rule '{rule}' provided to '{command}'", comment.pos
                    )

        if command == "tclint-disable":
            for rule in rules:
                # if in dictionary, already disabled - this has no effect
                if rule not in self._disable_regions:
                    self._disable_regions[rule] = comment.line
        elif command == "tclint-disable-line":
            line = comment.line
            self.ignore_lines[line].update(rules)
        elif command == "tclint-disable-next-line":
            line = comment.line + 1
            self.ignore_lines[line].update(rules)
        elif command == "tclint-enable":
            for rule in rules:
                if rule in self._disable_regions:
                    disable_start_line = self._disable_regions[rule]
                    disable_end_line = comment.line

                    for line in range(disable_start_line, disable_end_line + 1):
                        self.ignore_lines[line].add(rule)

                    del self._disable_regions[rule]
        else:
            self._warning(
                f"comment starts with '{command}', which looks like a tclint keyword."
                " Is this a typo?",
                comment.pos,
            )

    def _warning(self, message, pos):
        # TODO: formal warning mechanism
        prefix = self._path if self._path is not None else "(stdin)"
        print(f"Warning: {prefix}:{pos[0]}:{pos[1]}: {message}")



import re

from tclint.commands import get_commands
from tclint.violations import Rule, Violation

from tclint.syntax_tree import (
    Visitor,
    BracedExpression,
    Expression,
    BracedWord,
    QuotedWord,
    CommandSub,
)


class LineLengthChecker:
    """Ensures lines aren't too long.

    Reports 'line-length' violations.
    """

    # ref: https://github.com/eslint/eslint/blob/b29a16b22f234f6134475efb6c7be5ac946556ee/lib/rules/max-len.js#L101 # noqa: E501
    # ^ ironic lint waiver...
    URL_RE = re.compile(r"[^:/?#]:\/\/[^?#]")

    def check(self, input, _, config):
        violations = []
        for i, line in enumerate(input.split("\n")):
            if self.URL_RE.search(line) is not None:
                # ignore URLs
                continue

            lineno = i + 1
            if len(line) > config.style_line_length:
                start = (lineno, 1)
                end = (lineno, len(line) + 1)
                violations.append(
                    Violation(
                        Rule.LINE_LENGTH,
                        f"line length is {len(line)}, maximum allowed is"
                        f" {config.style_line_length}",
                        start,
                        end,
                    )
                )

        return violations


class TrailingWhitespaceChecker:
    """Ensures lines don't include trailing whitespace.

    Reports 'trailing-whitespace' violations.
    """

    def check(self, input, _, config):
        violations = []
        for i, line in enumerate(input.split("\n")):
            lineno = i + 1

            WHITESPACE = (" ", "\t")
            if line.endswith(WHITESPACE):
                start_col = len(line.rstrip("".join(WHITESPACE)))
                start = (lineno, start_col + 1)
                end = (lineno, len(line) + 1)
                violations.append(
                    Violation(
                        Rule.TRAILING_WHITESPACE,
                        "line has trailing whitespace",
                        start,
                        end,
                    )
                )

        return violations


class RedefinedBuiltinChecker(Visitor):
    """Ensures names of built-in commands aren't reused by proc definitions.

    Reports 'redefined-builtin' violations.
    """

    def check(self, _, tree, config):
        self._violations = []

        plugins = [config.commands] if config.commands is not None else []
        commands = get_commands(plugins)
        self._commands = commands.keys()

        tree.accept(self, recurse=True)

        return self._violations

    def visit_command(self, command):
        if command.routine.contents != "proc":
            return

        if len(command.args) == 0:
            # This is a syntax error, but should already be caught as a command-args
            # error by the parser's `proc` command handling.
            return

        name = command.args[0].contents

        if name in self._commands:
            self._violations.append(
                Violation(
                    Rule.REDEFINED_BUILTIN,
                    f"redefinition of built-in command '{name}'",
                    command.pos,
                    command.args[1].end_pos,
                )
            )


class UnbracedExprChecker(Visitor):
    def check(self, _, tree, __):
        self._violations = []
        tree.accept(self, recurse=True)
        return self._violations

    def visit_command(self, command):
        if command.routine.contents != "expr":
            return

        if len(command.args) == 0:
            # This is a syntax error, but should already be caught as a command-args
            # error by the parser's `expr` command handling.
            return

        if len(command.args) == 1 and isinstance(
            command.args[0], (BracedExpression, Expression)
        ):
            return

        # If we got here, tclint had trouble parsing the expression due to one of the
        # two following cases.

        for child in command.args:
            if child.contents is None:
                self._violations.append(
                    Violation(
                        Rule.UNBRACED_EXPR,
                        "expression with substitutions should be enclosed by braces",
                        command.args[0].pos,
                        command.args[-1].end_pos,
                    )
                )
                return

        for child in command.args:
            if isinstance(child, (BracedWord, QuotedWord)):
                self._violations.append(
                    Violation(
                        Rule.UNBRACED_EXPR,
                        "expression containing braced or quoted words should be"
                        " enclosed by braces",
                        command.args[0].pos,
                        command.args[-1].end_pos,
                    )
                )
                return

        # If we reach here, there's probably a bug in expr parsing logic.
        assert False, (
            "Children of expr node were different than expected, please file a bug"
            " report"
        )


class RedundantExprChecker(Visitor):
    def check(self, _, tree, __):
        self._violations = []
        tree.accept(self, recurse=True)
        return self._violations

    def _check_operand(self, operand):
        if not isinstance(operand, CommandSub) or len(operand.children) != 1:
            return

        command = operand.children[0]
        if command.routine.contents == "expr":
            self._violations.append(
                Violation(
                    Rule.REDUNDANT_EXPR,
                    "unnecessary command substitution within expression",
                    operand.pos,
                    operand.end_pos,
                )
            )

    def visit_braced_expression(self, expression):
        if len(expression.children) == 1:
            self._check_operand(expression.children[0])

    def visit_expression(self, expression):
        if len(expression.children) == 1:
            self._check_operand(expression.children[0])

    def visit_unary_op(self, expr):
        self._check_operand(expr.children[1])

    def visit_binary_op(self, expr):
        self._check_operand(expr.children[0])
        self._check_operand(expr.children[2])

    def visit_ternary_op(self, expr):
        self._check_operand(expr.children[0])
        self._check_operand(expr.children[2])
        self._check_operand(expr.children[4])

    def visit_function(self, function):
        for arg in function.children[1:]:
            self._check_operand(arg)


def get_checkers():
    checkers = (
        RedefinedBuiltinChecker(),
        UnbracedExprChecker(),
        RedundantExprChecker(),
        LineLengthChecker(),
        TrailingWhitespaceChecker(),
    )

    return checkers



import ply.lex as lex
from typing import Tuple

TOK_BACKSLASH_NEWLINE = "BACKSLASH_NEWLINE"
TOK_BACKSLASH_SUB = "BACKSLASH_SUB"
TOK_NEWLINE = "NEWLINE"
TOK_SEMI = "SEMI"
TOK_WS = "WS"
TOK_QUOTE = "QUOTE"
TOK_ARG_EXPANSION = "ARG_EXPANSION"
TOK_LBRACE = "LBRACE"
TOK_RBRACE = "RBRACE"
TOK_STAR = "STAR"
TOK_LBRACKET = "LBRACKET"
TOK_RBRACKET = "RBRACKET"
TOK_DOLLAR = "DOLLAR"
TOK_LPAREN = "LPAREN"
TOK_RPAREN = "RPAREN"
TOK_HASH = "HASH"
TOK_ALPHA_CHARS = "ALPHA_CHARS"
TOK_NUM_CHARS = "NUM_CHARS"
TOK_NAMESPACE_SEP = "NAMESPACE_SEP"
TOK_CHAR = "CHAR"
TOK_CONTENTS = "CONTENTS"
TOK_EOF = None

STATE_BRACEDWORD = "bracedword"


class TclSyntaxError(Exception):
    def __init__(self, message, start: Tuple[int, int], end: Tuple[int, int]):
        super().__init__(message)
        self.start = start
        self.end = end


class _LexTable:
    tokens = (
        TOK_BACKSLASH_NEWLINE,
        TOK_BACKSLASH_SUB,
        TOK_NEWLINE,
        TOK_SEMI,
        TOK_WS,
        TOK_QUOTE,
        TOK_ARG_EXPANSION,
        TOK_LBRACE,
        TOK_RBRACE,
        TOK_STAR,
        TOK_LBRACKET,
        TOK_RBRACKET,
        TOK_DOLLAR,
        TOK_LPAREN,
        TOK_RPAREN,
        TOK_HASH,
        TOK_ALPHA_CHARS,
        TOK_NUM_CHARS,
        TOK_NAMESPACE_SEP,
        TOK_CHAR,
        TOK_CONTENTS,
    )

    # This defines a conditional lexing state for parsing braced words. This is a
    # performance optimization; since there are few special characters in this context,
    # we can use a smaller set of tokens to parse them faster. This has a large impact
    # since most Tcl programs have a large number of braced words. Any token with
    # `bracedword` in its name is included in this state. Tokens that are included in
    # this state and the default state also include `INITIAL` in their name.
    states = ((STATE_BRACEDWORD, "exclusive"),)

    def _tok(self, t):
        pos = (t.lexer.lineno, t.lexer.colno)
        t.lexer.lineno += t.value.count("\n")
        index = t.value.rfind("\n")
        if index == -1:
            t.lexer.colno += len(t.value)
        else:
            remaining = t.value[index + 1 :]
            t.lexer.colno = len(remaining) + 1

        t.value = (t.value, pos)
        return t

    # Priority important
    def t_bracedword_INITIAL_BACKSLASH_NEWLINE(self, t):
        r"\\\n"
        return self._tok(t)

    # Priority important
    def t_bracedword_INITIAL_BACKSLASH_SUB(self, t):
        r"\\."
        return self._tok(t)

    def t_NEWLINE(self, t):
        r"\n"
        return self._tok(t)

    def t_SEMI(self, t):
        r";"
        return self._tok(t)

    # TODO: should use \s?
    def t_WS(self, t):
        r"[\t\v\f\r ]+"
        return self._tok(t)

    def t_QUOTE(self, t):
        r'"'
        return self._tok(t)

    # Must be higher priority than LBRACE
    def t_ARG_EXPANSION(self, t):
        r"\{\*\}"
        return self._tok(t)

    def t_bracedword_INITIAL_LBRACE(self, t):
        r"\{"
        return self._tok(t)

    def t_bracedword_INITIAL_RBRACE(self, t):
        r"\}"
        return self._tok(t)

    def t_STAR(self, t):
        r"\*"
        return self._tok(t)

    def t_LBRACKET(self, t):
        r"\["
        return self._tok(t)

    def t_RBRACKET(self, t):
        r"\]"
        return self._tok(t)

    def t_DOLLAR(self, t):
        r"\$"
        return self._tok(t)

    def t_LPAREN(self, t):
        r"\("
        return self._tok(t)

    def t_RPAREN(self, t):
        r"\)"
        return self._tok(t)

    def t_HASH(self, t):
        r"\#"
        return self._tok(t)

    # Valid non-numeric chars in variable names
    def t_ALPHA_CHARS(self, t):
        r"[A-Za-z_]+"
        return self._tok(t)

    # Valid numeric chars in variable names
    # This is split up from the above to facilitate expression parsing, since
    # e.g. 1eq1 can't be a single token.
    def t_NUM_CHARS(self, t):
        r"[0-9]+"
        return self._tok(t)

    def t_NAMESPACE_SEP(self, t):
        r"::+"
        return self._tok(t)

    def t_bracedword_CONTENTS(self, t):
        r"[^{}\\]+"
        return self._tok(t)

    # Catch-all. TODO: inefficient, should probably munch multiple chars
    def t_CHAR(self, t):
        r"."
        return self._tok(t)

    # Error handling rule
    # TODO: do we need this? since we have a catch-all...
    # there is a warning
    def t_bracedword_INITIAL_error(self, t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    def __init__(self):
        self.lexer = lex.lex(object=self)
        self.lexer.lineno = 1
        self.lexer.colno = 1

    def new_lexer(self, pos=None):
        lexer = self.lexer.clone()
        lexer.lineno = 1
        lexer.colno = 1

        if pos is not None:
            line, col = pos
            lexer.lineno = line
            lexer.colno = col

        return lexer


# Calling `lex.lex()` performs an expensive reflection process to generate the lexer.
# This singleton class holds a preinitialized lexer that can then be cloned to create
# individual instances.
LexTable = _LexTable()


class Lexer:
    def __init__(self, pos=None):
        self.lexer = LexTable.new_lexer(pos)
        self.current = None

    def input(self, text):
        self.lexer.input(text)
        self.current = self.lexer.token()

    def type(self):
        if self.current is None:
            return TOK_EOF
        return self.current.type

    def value(self):
        if self.current is None:
            return None
        return self.current.value[0]

    def pos(self):
        if self.current is None:
            return (self.lexer.lineno, self.lexer.colno)
        return self.current.value[1]

    def next(self):
        self.current = self.lexer.token()

    def expect(self, *tokens, message, pos):
        if self.type() not in tokens:
            self.next()  # munch another token to update position
            raise TclSyntaxError(message, pos, self.pos())

        self.next()

    def assert_(self, *tokens):
        assert self.current.type in tokens
        self.next()



from collections.abc import Callable
from voluptuous import Schema, Optional, Or, Self

# Need to define this as a Schema with required=True to ensure that this requirement
# persists through the Or in the main schema definition.
_command_args = Schema(
    {
        Optional("positionals", default=[]): [
            {
                "name": str,
                "required": bool,
                "value": Or({"type": "any"}, {"type": "variadic"}),
            }
        ],
        Optional("switches", default={}): {
            Optional(str): {
                "required": bool,
                "repeated": bool,
                "value": Or({"type": "any"}, None),
                Optional("metavar"): str,
            }
        },
    },
    required=True,
)

commands_schema = Schema(
    {Optional(str): Or(_command_args, None, {"subcommands": Self}, Callable)},
    required=True,
)

schema = Schema(
    {"name": str, "commands": commands_schema},
    required=True,
)



import json
from pathlib import Path

from tclint.cli.tclint import lint
from tclint.config import Config
from tclint.violations import Rule


def test_cmd_args_in_sub():
    """Ensure command args get checked in cmd sub."""
    script = "[puts]"
    violations = lint(script, Config(), Path())

    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_no_false_positive_arg_expansion():
    """Ensure that argument count checks don't flag when the number of arguments
    is ambiguous."""
    # $foo may be a list with two items, which is legal
    script = "rename {*}$foo"
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_accidental_return_expr():
    script = "return 5 + 2"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_ignore():
    script = "expr $foo"
    fake_path = Path("bad.tcl")
    violations = lint(script, Config(), fake_path)
    assert len(violations) == 1

    config = Config(ignore=[Rule("unbraced-expr")])
    violations = lint(script, config, fake_path)
    assert len(violations) == 0


def test_redefined_builtin():
    script = "proc puts {} {}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("redefined-builtin")
    assert violations[0].start == (1, 1)
    assert violations[0].end == (1, 13)


def test_redefined_builtin_plugin(tmp_path):
    script = r"proc foo {} {}"
    tmp_path = tmp_path / "plugin.json"
    with open(tmp_path, "w") as f:
        f.write(json.dumps({"name": "test", "commands": {"foo": None}}))

    violations = lint(script, Config(commands=tmp_path), Path())
    assert len(violations) == 1
    print(violations[0])
    assert violations[0].id == Rule("redefined-builtin")
    assert violations[0].start == (1, 1)
    assert violations[0].end == (1, 12)


def test_command_plugin(tmp_path):
    tmp_path = tmp_path / "plugin.json"
    with open(tmp_path, "w") as f:
        f.write(
            json.dumps({
                "name": "test",
                "commands": {
                    "foo": {
                        "positionals": [
                            {"name": "arg", "value": {"type": "any"}, "required": True}
                        ],
                    }
                },
            })
        )
    config = Config(commands=tmp_path)

    valid_script = "foo myarg"
    violations = lint(valid_script, config, Path())
    assert len(violations) == 0

    invalid_script = "foo"
    violations = lint(invalid_script, config, Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_no_violation_multiline_expr():
    script = r"""
if {1 && \
    2} {}"""
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_unbraced_expr_varsub():
    script = r"expr $foo"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 10)


def test_unbraced_expr_with_braced_word():
    script = r"expr 1 + {2}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 13)

    script = r'expr 1 + "2"'
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 13)

    script = r"expr {1} + {2}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 15)


def test_redundant_expr_check():
    script = r"""
expr {[expr 1]}
expr {([expr 1] + [expr 2])}
expr {[expr 1] ? [expr 2] : [expr 3]}
expr {max([expr 1], [expr 2])}
""".strip()

    violations = lint(script, Config(), Path())
    assert len(violations) == 8
    assert all(v.id == Rule.REDUNDANT_EXPR for v in violations)

    assert violations[0].start == (1, 7)
    assert violations[0].end == (1, 15)
    assert violations[1].start == (2, 8)
    assert violations[1].end == (2, 16)
    assert violations[2].start == (2, 19)
    assert violations[2].end == (2, 27)
    assert violations[3].start == (3, 7)
    assert violations[3].end == (3, 15)
    assert violations[4].start == (3, 18)
    assert violations[4].end == (3, 26)
    assert violations[5].start == (3, 29)
    assert violations[5].end == (3, 37)
    assert violations[6].start == (4, 11)
    assert violations[6].end == (4, 19)
    assert violations[7].start == (4, 21)
    assert violations[7].end == (4, 29)


def test_proc_args_bad():
    script = r"proc foo { { i 1 2 } } { }"
    violations = lint(script, Config(), Path())
    assert violations[0].id == Rule("command-args")


def test_expr_no_args():
    script = "expr"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_proc_no_args():
    script = "proc"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")



import pathlib
from typing import List, Dict, Union

from tclint.commands import builtin as _builtin
from tclint.commands.plugins import PluginManager

# import to expose in package
from tclint.commands.checks import CommandArgError

__all__ = ["CommandArgError", "validate_command_plugins", "get_commands"]


def validate_command_plugins(plugins: List[str]) -> List[str]:
    valid_plugins = []
    for plugin in set(plugins):
        if PluginManager.load(plugin) is not None:
            valid_plugins.append(plugin)

    return valid_plugins


def get_commands(plugins: List[Union[str, pathlib.Path]]) -> Dict:
    commands = {}
    commands.update(_builtin.commands)

    for plugin in plugins:
        if isinstance(plugin, str):
            plugin_commands = PluginManager.load(plugin)
        elif isinstance(plugin, pathlib.Path):
            plugin_commands = PluginManager.load_from_spec(plugin)
        else:
            raise TypeError(f"Plugins must be strings or paths, got {type(plugin)}")

        if plugin_commands is not None:
            commands.update(plugin_commands)

    return commands
